# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..weight import Weight
from ..three_item import ThreeItem
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ItemGetServerPacket(Packet):
    """
    Reply to taking items from the ground
    """
    _byte_size: int = 0
    _taken_item_index: int
    _taken_item: ThreeItem
    _weight: Weight

    def __init__(self, *, taken_item_index: int, taken_item: ThreeItem, weight: Weight):
        """
        Create a new instance of ItemGetServerPacket.

        Args:
            taken_item_index (int): (Value range is 0-64008.)
            taken_item (ThreeItem): 
            weight (Weight): 
        """
        self._taken_item_index = taken_item_index
        self._taken_item = taken_item
        self._weight = weight

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def taken_item_index(self) -> int:
        return self._taken_item_index

    @property
    def taken_item(self) -> ThreeItem:
        return self._taken_item

    @property
    def weight(self) -> Weight:
        return self._weight

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Item

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Get

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        ItemGetServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "ItemGetServerPacket") -> None:
        """
        Serializes an instance of `ItemGetServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ItemGetServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._taken_item_index is None:
                raise SerializationError("taken_item_index must be provided.")
            writer.add_short(data._taken_item_index)
            if data._taken_item is None:
                raise SerializationError("taken_item must be provided.")
            ThreeItem.serialize(writer, data._taken_item)
            if data._weight is None:
                raise SerializationError("weight must be provided.")
            Weight.serialize(writer, data._weight)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ItemGetServerPacket":
        """
        Deserializes an instance of `ItemGetServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ItemGetServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            taken_item_index = reader.get_short()
            taken_item = ThreeItem.deserialize(reader)
            weight = Weight.deserialize(reader)
            result = ItemGetServerPacket(taken_item_index=taken_item_index, taken_item=taken_item, weight=weight)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ItemGetServerPacket(byte_size={repr(self._byte_size)}, taken_item_index={repr(self._taken_item_index)}, taken_item={repr(self._taken_item)}, weight={repr(self._weight)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .item_type import ItemType
from .item_subtype import ItemSubtype
from .item_special import ItemSpecial
from .item_size import ItemSize
from .element import Element
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class EifRecord:
    """
    Record of Item data in an Endless Item File
    """
    _byte_size: int = 0
    _name_length: int
    _name: str
    _graphic_id: int
    _type: ItemType
    _subtype: ItemSubtype
    _special: ItemSpecial
    _hp: int
    _tp: int
    _min_damage: int
    _max_damage: int
    _accuracy: int
    _evade: int
    _armor: int
    _return_damage: int
    _str: int
    _intl: int
    _wis: int
    _agi: int
    _con: int
    _cha: int
    _light_resistance: int
    _dark_resistance: int
    _earth_resistance: int
    _air_resistance: int
    _water_resistance: int
    _fire_resistance: int
    _spec1: int
    _spec2: int
    _spec3: int
    _level_requirement: int
    _class_requirement: int
    _str_requirement: int
    _int_requirement: int
    _wis_requirement: int
    _agi_requirement: int
    _con_requirement: int
    _cha_requirement: int
    _element: Element
    _element_damage: int
    _weight: int
    _size: ItemSize

    def __init__(self, *, name: str, graphic_id: int, type: ItemType, subtype: ItemSubtype, special: ItemSpecial, hp: int, tp: int, min_damage: int, max_damage: int, accuracy: int, evade: int, armor: int, return_damage: int, str: int, intl: int, wis: int, agi: int, con: int, cha: int, light_resistance: int, dark_resistance: int, earth_resistance: int, air_resistance: int, water_resistance: int, fire_resistance: int, spec1: int, spec2: int, spec3: int, level_requirement: int, class_requirement: int, str_requirement: int, int_requirement: int, wis_requirement: int, agi_requirement: int, con_requirement: int, cha_requirement: int, element: Element, element_damage: int, weight: int, size: ItemSize):
        """
        Create a new instance of EifRecord.

        Args:
            name (str): (Length must be 252 or less.)
            graphic_id (int): (Value range is 0-64008.)
            type (ItemType): 
            subtype (ItemSubtype): 
            special (ItemSpecial): 
            hp (int): (Value range is 0-64008.)
            tp (int): (Value range is 0-64008.)
            min_damage (int): (Value range is 0-64008.)
            max_damage (int): (Value range is 0-64008.)
            accuracy (int): (Value range is 0-64008.)
            evade (int): (Value range is 0-64008.)
            armor (int): (Value range is 0-64008.)
            return_damage (int): (Value range is 0-252.)
            str (int): (Value range is 0-252.)
            intl (int): (Value range is 0-252.)
            wis (int): (Value range is 0-252.)
            agi (int): (Value range is 0-252.)
            con (int): (Value range is 0-252.)
            cha (int): (Value range is 0-252.)
            light_resistance (int): (Value range is 0-252.)
            dark_resistance (int): (Value range is 0-252.)
            earth_resistance (int): (Value range is 0-252.)
            air_resistance (int): (Value range is 0-252.)
            water_resistance (int): (Value range is 0-252.)
            fire_resistance (int): (Value range is 0-252.)
            spec1 (int): Holds one the following values, depending on item type: scroll_map, doll_graphic, exp_reward, hair_color, effect, key, alcohol_potency (Value range is 0-16194276.)
            spec2 (int): Holds one the following values, depending on item type: scroll_x, gender (Value range is 0-252.)
            spec3 (int): Holds one the following values, depending on item type: scroll_y (Value range is 0-252.)
            level_requirement (int): (Value range is 0-64008.)
            class_requirement (int): (Value range is 0-64008.)
            str_requirement (int): (Value range is 0-64008.)
            int_requirement (int): (Value range is 0-64008.)
            wis_requirement (int): (Value range is 0-64008.)
            agi_requirement (int): (Value range is 0-64008.)
            con_requirement (int): (Value range is 0-64008.)
            cha_requirement (int): (Value range is 0-64008.)
            element (Element): 
            element_damage (int): (Value range is 0-252.)
            weight (int): (Value range is 0-252.)
            size (ItemSize): 
        """
        self._name = name
        self._name_length = len(self._name)
        self._graphic_id = graphic_id
        self._type = type
        self._subtype = subtype
        self._special = special
        self._hp = hp
        self._tp = tp
        self._min_damage = min_damage
        self._max_damage = max_damage
        self._accuracy = accuracy
        self._evade = evade
        self._armor = armor
        self._return_damage = return_damage
        self._str = str
        self._intl = intl
        self._wis = wis
        self._agi = agi
        self._con = con
        self._cha = cha
        self._light_resistance = light_resistance
        self._dark_resistance = dark_resistance
        self._earth_resistance = earth_resistance
        self._air_resistance = air_resistance
        self._water_resistance = water_resistance
        self._fire_resistance = fire_resistance
        self._spec1 = spec1
        self._spec2 = spec2
        self._spec3 = spec3
        self._level_requirement = level_requirement
        self._class_requirement = class_requirement
        self._str_requirement = str_requirement
        self._int_requirement = int_requirement
        self._wis_requirement = wis_requirement
        self._agi_requirement = agi_requirement
        self._con_requirement = con_requirement
        self._cha_requirement = cha_requirement
        self._element = element
        self._element_damage = element_damage
        self._weight = weight
        self._size = size

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def graphic_id(self) -> int:
        return self._graphic_id

    @property
    def type(self) -> ItemType:
        return self._type

    @property
    def subtype(self) -> ItemSubtype:
        return self._subtype

    @property
    def special(self) -> ItemSpecial:
        return self._special

    @property
    def hp(self) -> int:
        return self._hp

    @property
    def tp(self) -> int:
        return self._tp

    @property
    def min_damage(self) -> int:
        return self._min_damage

    @property
    def max_damage(self) -> int:
        return self._max_damage

    @property
    def accuracy(self) -> int:
        return self._accuracy

    @property
    def evade(self) -> int:
        return self._evade

    @property
    def armor(self) -> int:
        return self._armor

    @property
    def return_damage(self) -> int:
        return self._return_damage

    @property
    def str(self) -> int:
        return self._str

    @property
    def intl(self) -> int:
        return self._intl

    @property
    def wis(self) -> int:
        return self._wis

    @property
    def agi(self) -> int:
        return self._agi

    @property
    def con(self) -> int:
        return self._con

    @property
    def cha(self) -> int:
        return self._cha

    @property
    def light_resistance(self) -> int:
        return self._light_resistance

    @property
    def dark_resistance(self) -> int:
        return self._dark_resistance

    @property
    def earth_resistance(self) -> int:
        return self._earth_resistance

    @property
    def air_resistance(self) -> int:
        return self._air_resistance

    @property
    def water_resistance(self) -> int:
        return self._water_resistance

    @property
    def fire_resistance(self) -> int:
        return self._fire_resistance

    @property
    def spec1(self) -> int:
        """
        Holds one the following values, depending on item type:
        scroll_map, doll_graphic, exp_reward, hair_color, effect, key, alcohol_potency
        """
        return self._spec1

    @property
    def spec2(self) -> int:
        """
        Holds one the following values, depending on item type:
        scroll_x, gender
        """
        return self._spec2

    @property
    def spec3(self) -> int:
        """
        Holds one the following values, depending on item type:
        scroll_y
        """
        return self._spec3

    @property
    def level_requirement(self) -> int:
        return self._level_requirement

    @property
    def class_requirement(self) -> int:
        return self._class_requirement

    @property
    def str_requirement(self) -> int:
        return self._str_requirement

    @property
    def int_requirement(self) -> int:
        return self._int_requirement

    @property
    def wis_requirement(self) -> int:
        return self._wis_requirement

    @property
    def agi_requirement(self) -> int:
        return self._agi_requirement

    @property
    def con_requirement(self) -> int:
        return self._con_requirement

    @property
    def cha_requirement(self) -> int:
        return self._cha_requirement

    @property
    def element(self) -> Element:
        return self._element

    @property
    def element_damage(self) -> int:
        return self._element_damage

    @property
    def weight(self) -> int:
        return self._weight

    @property
    def size(self) -> ItemSize:
        return self._size

    @staticmethod
    def serialize(writer: EoWriter, data: "EifRecord") -> None:
        """
        Serializes an instance of `EifRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (EifRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._name_length is None:
                raise SerializationError("name_length must be provided.")
            writer.add_char(data._name_length)
            if data._name is None:
                raise SerializationError("name must be provided.")
            if len(data._name) > 252:
                raise SerializationError(f"Expected length of name to be 252 or less, got {len(data._name)}.")
            writer.add_fixed_string(data._name, data._name_length, False)
            if data._graphic_id is None:
                raise SerializationError("graphic_id must be provided.")
            writer.add_short(data._graphic_id)
            if data._type is None:
                raise SerializationError("type must be provided.")
            writer.add_char(int(data._type))
            if data._subtype is None:
                raise SerializationError("subtype must be provided.")
            writer.add_char(int(data._subtype))
            if data._special is None:
                raise SerializationError("special must be provided.")
            writer.add_char(int(data._special))
            if data._hp is None:
                raise SerializationError("hp must be provided.")
            writer.add_short(data._hp)
            if data._tp is None:
                raise SerializationError("tp must be provided.")
            writer.add_short(data._tp)
            if data._min_damage is None:
                raise SerializationError("min_damage must be provided.")
            writer.add_short(data._min_damage)
            if data._max_damage is None:
                raise SerializationError("max_damage must be provided.")
            writer.add_short(data._max_damage)
            if data._accuracy is None:
                raise SerializationError("accuracy must be provided.")
            writer.add_short(data._accuracy)
            if data._evade is None:
                raise SerializationError("evade must be provided.")
            writer.add_short(data._evade)
            if data._armor is None:
                raise SerializationError("armor must be provided.")
            writer.add_short(data._armor)
            if data._return_damage is None:
                raise SerializationError("return_damage must be provided.")
            writer.add_char(data._return_damage)
            if data._str is None:
                raise SerializationError("str must be provided.")
            writer.add_char(data._str)
            if data._intl is None:
                raise SerializationError("intl must be provided.")
            writer.add_char(data._intl)
            if data._wis is None:
                raise SerializationError("wis must be provided.")
            writer.add_char(data._wis)
            if data._agi is None:
                raise SerializationError("agi must be provided.")
            writer.add_char(data._agi)
            if data._con is None:
                raise SerializationError("con must be provided.")
            writer.add_char(data._con)
            if data._cha is None:
                raise SerializationError("cha must be provided.")
            writer.add_char(data._cha)
            if data._light_resistance is None:
                raise SerializationError("light_resistance must be provided.")
            writer.add_char(data._light_resistance)
            if data._dark_resistance is None:
                raise SerializationError("dark_resistance must be provided.")
            writer.add_char(data._dark_resistance)
            if data._earth_resistance is None:
                raise SerializationError("earth_resistance must be provided.")
            writer.add_char(data._earth_resistance)
            if data._air_resistance is None:
                raise SerializationError("air_resistance must be provided.")
            writer.add_char(data._air_resistance)
            if data._water_resistance is None:
                raise SerializationError("water_resistance must be provided.")
            writer.add_char(data._water_resistance)
            if data._fire_resistance is None:
                raise SerializationError("fire_resistance must be provided.")
            writer.add_char(data._fire_resistance)
            if data._spec1 is None:
                raise SerializationError("spec1 must be provided.")
            writer.add_three(data._spec1)
            if data._spec2 is None:
                raise SerializationError("spec2 must be provided.")
            writer.add_char(data._spec2)
            if data._spec3 is None:
                raise SerializationError("spec3 must be provided.")
            writer.add_char(data._spec3)
            if data._level_requirement is None:
                raise SerializationError("level_requirement must be provided.")
            writer.add_short(data._level_requirement)
            if data._class_requirement is None:
                raise SerializationError("class_requirement must be provided.")
            writer.add_short(data._class_requirement)
            if data._str_requirement is None:
                raise SerializationError("str_requirement must be provided.")
            writer.add_short(data._str_requirement)
            if data._int_requirement is None:
                raise SerializationError("int_requirement must be provided.")
            writer.add_short(data._int_requirement)
            if data._wis_requirement is None:
                raise SerializationError("wis_requirement must be provided.")
            writer.add_short(data._wis_requirement)
            if data._agi_requirement is None:
                raise SerializationError("agi_requirement must be provided.")
            writer.add_short(data._agi_requirement)
            if data._con_requirement is None:
                raise SerializationError("con_requirement must be provided.")
            writer.add_short(data._con_requirement)
            if data._cha_requirement is None:
                raise SerializationError("cha_requirement must be provided.")
            writer.add_short(data._cha_requirement)
            if data._element is None:
                raise SerializationError("element must be provided.")
            writer.add_char(int(data._element))
            if data._element_damage is None:
                raise SerializationError("element_damage must be provided.")
            writer.add_char(data._element_damage)
            if data._weight is None:
                raise SerializationError("weight must be provided.")
            writer.add_char(data._weight)
            writer.add_char(0)
            if data._size is None:
                raise SerializationError("size must be provided.")
            writer.add_char(int(data._size))
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "EifRecord":
        """
        Deserializes an instance of `EifRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            EifRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            name_length = reader.get_char()
            name = reader.get_fixed_string(name_length, False)
            graphic_id = reader.get_short()
            type = ItemType(reader.get_char())
            subtype = ItemSubtype(reader.get_char())
            special = ItemSpecial(reader.get_char())
            hp = reader.get_short()
            tp = reader.get_short()
            min_damage = reader.get_short()
            max_damage = reader.get_short()
            accuracy = reader.get_short()
            evade = reader.get_short()
            armor = reader.get_short()
            return_damage = reader.get_char()
            str = reader.get_char()
            intl = reader.get_char()
            wis = reader.get_char()
            agi = reader.get_char()
            con = reader.get_char()
            cha = reader.get_char()
            light_resistance = reader.get_char()
            dark_resistance = reader.get_char()
            earth_resistance = reader.get_char()
            air_resistance = reader.get_char()
            water_resistance = reader.get_char()
            fire_resistance = reader.get_char()
            spec1 = reader.get_three()
            spec2 = reader.get_char()
            spec3 = reader.get_char()
            level_requirement = reader.get_short()
            class_requirement = reader.get_short()
            str_requirement = reader.get_short()
            int_requirement = reader.get_short()
            wis_requirement = reader.get_short()
            agi_requirement = reader.get_short()
            con_requirement = reader.get_short()
            cha_requirement = reader.get_short()
            element = Element(reader.get_char())
            element_damage = reader.get_char()
            weight = reader.get_char()
            reader.get_char()
            size = ItemSize(reader.get_char())
            result = EifRecord(name=name, graphic_id=graphic_id, type=type, subtype=subtype, special=special, hp=hp, tp=tp, min_damage=min_damage, max_damage=max_damage, accuracy=accuracy, evade=evade, armor=armor, return_damage=return_damage, str=str, intl=intl, wis=wis, agi=agi, con=con, cha=cha, light_resistance=light_resistance, dark_resistance=dark_resistance, earth_resistance=earth_resistance, air_resistance=air_resistance, water_resistance=water_resistance, fire_resistance=fire_resistance, spec1=spec1, spec2=spec2, spec3=spec3, level_requirement=level_requirement, class_requirement=class_requirement, str_requirement=str_requirement, int_requirement=int_requirement, wis_requirement=wis_requirement, agi_requirement=agi_requirement, con_requirement=con_requirement, cha_requirement=cha_requirement, element=element, element_damage=element_damage, weight=weight, size=size)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"EifRecord(byte_size={repr(self._byte_size)}, name={repr(self._name)}, graphic_id={repr(self._graphic_id)}, type={repr(self._type)}, subtype={repr(self._subtype)}, special={repr(self._special)}, hp={repr(self._hp)}, tp={repr(self._tp)}, min_damage={repr(self._min_damage)}, max_damage={repr(self._max_damage)}, accuracy={repr(self._accuracy)}, evade={repr(self._evade)}, armor={repr(self._armor)}, return_damage={repr(self._return_damage)}, str={repr(self._str)}, intl={repr(self._intl)}, wis={repr(self._wis)}, agi={repr(self._agi)}, con={repr(self._con)}, cha={repr(self._cha)}, light_resistance={repr(self._light_resistance)}, dark_resistance={repr(self._dark_resistance)}, earth_resistance={repr(self._earth_resistance)}, air_resistance={repr(self._air_resistance)}, water_resistance={repr(self._water_resistance)}, fire_resistance={repr(self._fire_resistance)}, spec1={repr(self._spec1)}, spec2={repr(self._spec2)}, spec3={repr(self._spec3)}, level_requirement={repr(self._level_requirement)}, class_requirement={repr(self._class_requirement)}, str_requirement={repr(self._str_requirement)}, int_requirement={repr(self._int_requirement)}, wis_requirement={repr(self._wis_requirement)}, agi_requirement={repr(self._agi_requirement)}, con_requirement={repr(self._con_requirement)}, cha_requirement={repr(self._cha_requirement)}, element={repr(self._element)}, element_damage={repr(self._element_damage)}, weight={repr(self._weight)}, size={repr(self._size)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class SpellTargetType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Target type of a spell cast
    """
    Player = 1
    Npc = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from typing import Union
from .file_type import FileType
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class WelcomeAgreeClientPacket(Packet):
    """
    Requesting a file
    """
    _byte_size: int = 0
    _file_type: FileType
    _session_id: int
    _file_type_data: 'WelcomeAgreeClientPacket.FileTypeData'

    def __init__(self, *, file_type: FileType, session_id: int, file_type_data: 'WelcomeAgreeClientPacket.FileTypeData' = None):
        """
        Create a new instance of WelcomeAgreeClientPacket.

        Args:
            file_type (FileType): 
            session_id (int): (Value range is 0-64008.)
            file_type_data (WelcomeAgreeClientPacket.FileTypeData): Data associated with the `file_type` field.
        """
        self._file_type = file_type
        self._session_id = session_id
        self._file_type_data = file_type_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def file_type(self) -> FileType:
        return self._file_type

    @property
    def session_id(self) -> int:
        return self._session_id

    @property
    def file_type_data(self) -> 'WelcomeAgreeClientPacket.FileTypeData':
        """
        WelcomeAgreeClientPacket.FileTypeData: Data associated with the `file_type` field.
        """
        return self._file_type_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Welcome

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Agree

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        WelcomeAgreeClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket") -> None:
        """
        Serializes an instance of `WelcomeAgreeClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (WelcomeAgreeClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._file_type is None:
                raise SerializationError("file_type must be provided.")
            writer.add_char(int(data._file_type))
            if data._session_id is None:
                raise SerializationError("session_id must be provided.")
            writer.add_short(data._session_id)
            if data._file_type == FileType.Emf:
                if not isinstance(data._file_type_data, WelcomeAgreeClientPacket.FileTypeDataEmf):
                    raise SerializationError("Expected file_type_data to be type WelcomeAgreeClientPacket.FileTypeDataEmf for file_type " + FileType(data._file_type).name + ".")
                WelcomeAgreeClientPacket.FileTypeDataEmf.serialize(writer, data._file_type_data)
            elif data._file_type == FileType.Eif:
                if not isinstance(data._file_type_data, WelcomeAgreeClientPacket.FileTypeDataEif):
                    raise SerializationError("Expected file_type_data to be type WelcomeAgreeClientPacket.FileTypeDataEif for file_type " + FileType(data._file_type).name + ".")
                WelcomeAgreeClientPacket.FileTypeDataEif.serialize(writer, data._file_type_data)
            elif data._file_type == FileType.Enf:
                if not isinstance(data._file_type_data, WelcomeAgreeClientPacket.FileTypeDataEnf):
                    raise SerializationError("Expected file_type_data to be type WelcomeAgreeClientPacket.FileTypeDataEnf for file_type " + FileType(data._file_type).name + ".")
                WelcomeAgreeClientPacket.FileTypeDataEnf.serialize(writer, data._file_type_data)
            elif data._file_type == FileType.Esf:
                if not isinstance(data._file_type_data, WelcomeAgreeClientPacket.FileTypeDataEsf):
                    raise SerializationError("Expected file_type_data to be type WelcomeAgreeClientPacket.FileTypeDataEsf for file_type " + FileType(data._file_type).name + ".")
                WelcomeAgreeClientPacket.FileTypeDataEsf.serialize(writer, data._file_type_data)
            elif data._file_type == FileType.Ecf:
                if not isinstance(data._file_type_data, WelcomeAgreeClientPacket.FileTypeDataEcf):
                    raise SerializationError("Expected file_type_data to be type WelcomeAgreeClientPacket.FileTypeDataEcf for file_type " + FileType(data._file_type).name + ".")
                WelcomeAgreeClientPacket.FileTypeDataEcf.serialize(writer, data._file_type_data)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket":
        """
        Deserializes an instance of `WelcomeAgreeClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            WelcomeAgreeClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            file_type = FileType(reader.get_char())
            session_id = reader.get_short()
            file_type_data: WelcomeAgreeClientPacket.FileTypeData = None
            if file_type == FileType.Emf:
                file_type_data = WelcomeAgreeClientPacket.FileTypeDataEmf.deserialize(reader)
            elif file_type == FileType.Eif:
                file_type_data = WelcomeAgreeClientPacket.FileTypeDataEif.deserialize(reader)
            elif file_type == FileType.Enf:
                file_type_data = WelcomeAgreeClientPacket.FileTypeDataEnf.deserialize(reader)
            elif file_type == FileType.Esf:
                file_type_data = WelcomeAgreeClientPacket.FileTypeDataEsf.deserialize(reader)
            elif file_type == FileType.Ecf:
                file_type_data = WelcomeAgreeClientPacket.FileTypeDataEcf.deserialize(reader)
            result = WelcomeAgreeClientPacket(file_type=file_type, session_id=session_id, file_type_data=file_type_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"WelcomeAgreeClientPacket(byte_size={repr(self._byte_size)}, file_type={repr(self._file_type)}, session_id={repr(self._session_id)}, file_type_data={repr(self._file_type_data)})"

    FileTypeData = Union['WelcomeAgreeClientPacket.FileTypeDataEmf', 'WelcomeAgreeClientPacket.FileTypeDataEif', 'WelcomeAgreeClientPacket.FileTypeDataEnf', 'WelcomeAgreeClientPacket.FileTypeDataEsf', 'WelcomeAgreeClientPacket.FileTypeDataEcf', None]
    """
    Data associated with different values of the `file_type` field.
    """

    class FileTypeDataEmf:
        """
        Data associated with file_type value FileType.Emf
        """
        _byte_size: int = 0
        _file_id: int

        def __init__(self, *, file_id: int):
            """
            Create a new instance of WelcomeAgreeClientPacket.FileTypeDataEmf.

            Args:
                file_id (int): (Value range is 0-64008.)
            """
            self._file_id = file_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def file_id(self) -> int:
            return self._file_id

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket.FileTypeDataEmf") -> None:
            """
            Serializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEmf` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeAgreeClientPacket.FileTypeDataEmf): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._file_id is None:
                    raise SerializationError("file_id must be provided.")
                writer.add_short(data._file_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket.FileTypeDataEmf":
            """
            Deserializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEmf` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeAgreeClientPacket.FileTypeDataEmf: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                file_id = reader.get_short()
                result = WelcomeAgreeClientPacket.FileTypeDataEmf(file_id=file_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeAgreeClientPacket.FileTypeDataEmf(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)})"

    class FileTypeDataEif:
        """
        Data associated with file_type value FileType.Eif
        """
        _byte_size: int = 0
        _file_id: int

        def __init__(self, *, file_id: int):
            """
            Create a new instance of WelcomeAgreeClientPacket.FileTypeDataEif.

            Args:
                file_id (int): (Value range is 0-252.)
            """
            self._file_id = file_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def file_id(self) -> int:
            return self._file_id

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket.FileTypeDataEif") -> None:
            """
            Serializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEif` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeAgreeClientPacket.FileTypeDataEif): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._file_id is None:
                    raise SerializationError("file_id must be provided.")
                writer.add_char(data._file_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket.FileTypeDataEif":
            """
            Deserializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEif` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeAgreeClientPacket.FileTypeDataEif: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                file_id = reader.get_char()
                result = WelcomeAgreeClientPacket.FileTypeDataEif(file_id=file_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeAgreeClientPacket.FileTypeDataEif(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)})"

    class FileTypeDataEnf:
        """
        Data associated with file_type value FileType.Enf
        """
        _byte_size: int = 0
        _file_id: int

        def __init__(self, *, file_id: int):
            """
            Create a new instance of WelcomeAgreeClientPacket.FileTypeDataEnf.

            Args:
                file_id (int): (Value range is 0-252.)
            """
            self._file_id = file_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def file_id(self) -> int:
            return self._file_id

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket.FileTypeDataEnf") -> None:
            """
            Serializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEnf` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeAgreeClientPacket.FileTypeDataEnf): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._file_id is None:
                    raise SerializationError("file_id must be provided.")
                writer.add_char(data._file_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket.FileTypeDataEnf":
            """
            Deserializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEnf` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeAgreeClientPacket.FileTypeDataEnf: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                file_id = reader.get_char()
                result = WelcomeAgreeClientPacket.FileTypeDataEnf(file_id=file_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeAgreeClientPacket.FileTypeDataEnf(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)})"

    class FileTypeDataEsf:
        """
        Data associated with file_type value FileType.Esf
        """
        _byte_size: int = 0
        _file_id: int

        def __init__(self, *, file_id: int):
            """
            Create a new instance of WelcomeAgreeClientPacket.FileTypeDataEsf.

            Args:
                file_id (int): (Value range is 0-252.)
            """
            self._file_id = file_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def file_id(self) -> int:
            return self._file_id

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket.FileTypeDataEsf") -> None:
            """
            Serializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEsf` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeAgreeClientPacket.FileTypeDataEsf): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._file_id is None:
                    raise SerializationError("file_id must be provided.")
                writer.add_char(data._file_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket.FileTypeDataEsf":
            """
            Deserializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEsf` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeAgreeClientPacket.FileTypeDataEsf: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                file_id = reader.get_char()
                result = WelcomeAgreeClientPacket.FileTypeDataEsf(file_id=file_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeAgreeClientPacket.FileTypeDataEsf(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)})"

    class FileTypeDataEcf:
        """
        Data associated with file_type value FileType.Ecf
        """
        _byte_size: int = 0
        _file_id: int

        def __init__(self, *, file_id: int):
            """
            Create a new instance of WelcomeAgreeClientPacket.FileTypeDataEcf.

            Args:
                file_id (int): (Value range is 0-252.)
            """
            self._file_id = file_id

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def file_id(self) -> int:
            return self._file_id

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeAgreeClientPacket.FileTypeDataEcf") -> None:
            """
            Serializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEcf` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeAgreeClientPacket.FileTypeDataEcf): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._file_id is None:
                    raise SerializationError("file_id must be provided.")
                writer.add_char(data._file_id)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeAgreeClientPacket.FileTypeDataEcf":
            """
            Deserializes an instance of `WelcomeAgreeClientPacket.FileTypeDataEcf` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeAgreeClientPacket.FileTypeDataEcf: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                file_id = reader.get_char()
                result = WelcomeAgreeClientPacket.FileTypeDataEcf(file_id=file_id)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeAgreeClientPacket.FileTypeDataEcf(byte_size={repr(self._byte_size)}, file_id={repr(self._file_id)})"



import os
from pathlib import Path

from protocol_code_generator.generate.code_block import CodeBlock


class PythonFile:
    def __init__(self, relative_path, code_block, *, module_docstring=None):
        self._relative_path = relative_path
        self._code_block = code_block
        self._module_docstring = module_docstring

    def write(self, root_path):
        output_path = os.path.join(root_path, self._relative_path)

        header = CodeBlock()
        header.add_line("# Generated from the eo-protocol XML specification.")
        header.add_line("#")
        header.add_line("# This file should not be modified.")
        header.add_line("# Changes will be lost when code is regenerated.")
        header.add_line()

        if self._module_docstring is not None:
            header.add_code_block(self._module_docstring)
            header.add_line()

        os.makedirs(os.path.dirname(output_path), exist_ok=True)

        package_path = os.path.dirname(self._relative_path)
        package_path = os.path.join("eolib/protocol/_generated", package_path)
        package_path = Path(package_path).as_posix()
        package_path = package_path.replace('/', '.')

        with open(output_path, "w", encoding="utf-8") as file:
            file.write(header.to_string(package_path) + self._code_block.to_string(package_path))

    @property
    def relative_path(self):
        return self._relative_path



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class MarriageRequestType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Request type sent with MARRIAGE_REQUEST packet
    """
    MarriageApproval = 1
    Divorce = 2



import os
from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class ProtocolBuildHook(BuildHookInterface):
    def clean(self, versions: list) -> None:
        self._run_python_script("./protocol.py clean")

    def initialize(self, version: str, build_data: dict) -> None:
        self._run_python_script("./protocol.py generate")

    def _run_python_script(self, python_args: str) -> None:
        command = f"python {python_args}"
        code: int = os.system(command)
        if code != 0:
            raise RuntimeError(f'Command "{command}" failed with exit code {code}.')



from protocol_code_generator.type.blob_type import BlobType
from protocol_code_generator.type.bool_type import BoolType
from protocol_code_generator.type.enum_type import EnumType, EnumValue
from protocol_code_generator.type.has_underlying_type import HasUnderlyingType
from protocol_code_generator.type.integer_type import IntegerType
from protocol_code_generator.type.length import Length
from protocol_code_generator.type.string_type import StringType
from protocol_code_generator.type.struct_type import StructType
from protocol_code_generator.util.number_utils import try_parse_int
from protocol_code_generator.util.xml_utils import (
    get_instructions,
    get_required_string_attribute,
    get_string_attribute,
    get_text,
)


class TypeFactory:
    def __init__(self):
        self.unresolved_types = {}
        self.types = {}

    def get_type(self, name: str, length=None):
        if not length:
            length = Length.unspecified()
        if length.specified:
            return TypeFactory._create_type_with_specified_length(name, length)
        if name not in self.types:
            self.types[name] = self._create_type(name, length)
        return self.types[name]

    def define_custom_type(self, protocol_type, source_path):
        name = get_required_string_attribute(protocol_type, "name")
        if name in self.unresolved_types:
            return False
        self.unresolved_types[name] = UnresolvedCustomType(protocol_type, source_path)
        return True

    def clear(self):
        self.unresolved_types.clear()
        self.types.clear()

    def _create_type(self, name, length):
        underlying_type = self._read_underlying_type(name)
        if underlying_type is not None:
            name = name[: name.index(":")]

        result = None

        if name in ["byte", "char"]:
            result = IntegerType(name, 1)
        elif name == "short":
            result = IntegerType(name, 2)
        elif name == "three":
            result = IntegerType(name, 3)
        elif name == "int":
            result = IntegerType(name, 4)
        elif name == "bool":
            if underlying_type is None:
                underlying_type = self.get_type("char")
            result = BoolType(underlying_type)
        elif name in ["string", "encoded_string"]:
            result = StringType(name, length)
        elif name == "blob":
            result = BlobType()
        else:
            result = self._create_custom_type(name, underlying_type)

        if underlying_type is not None and not isinstance(result, HasUnderlyingType):
            raise RuntimeError(
                f"{result.name} has no underlying type, so {underlying_type.name} is not allowed "
                + "as an underlying type override."
            )

        return result

    def _read_underlying_type(self, name):
        parts = name.split(":")

        if len(parts) == 1:
            return None
        elif len(parts) == 2:
            type_name, underlying_type_name = parts
            if type_name == underlying_type_name:
                raise RuntimeError(f"{type_name} type cannot specify itself as an underlying type.")
            underlying_type = self.get_type(underlying_type_name)
            if not isinstance(underlying_type, IntegerType):
                raise RuntimeError(
                    f"{underlying_type.name} is not a numeric type, so it cannot be specified as "
                    + "an underlying type."
                )
            return underlying_type
        else:
            raise RuntimeError(f'"{name}" type syntax is invalid. (Only one colon is allowed)')

    def _create_custom_type(self, name, underlying_type_override):
        unresolved_type = self.unresolved_types.get(name)
        if not unresolved_type:
            raise RuntimeError(f"{name} type is not defined.")

        if unresolved_type.type_xml.tag == "enum":
            return self._create_enum_type(
                unresolved_type.type_xml, underlying_type_override, unresolved_type.relative_path
            )
        elif unresolved_type.type_xml.tag == "struct":
            return self._create_struct_type(unresolved_type.type_xml, unresolved_type.relative_path)
        else:
            raise RuntimeError(
                f'Unhandled CustomType xml element: <{unresolved_type.type_xml.tag}>'
            )

    def _create_enum_type(self, protocol_enum, underlying_type_override, relative_path):
        underlying_type = underlying_type_override
        enum_name = get_required_string_attribute(protocol_enum, "name")

        if underlying_type is None:
            underlying_type_name = get_required_string_attribute(protocol_enum, "type")
            if enum_name == underlying_type_name:
                raise RuntimeError(f"{enum_name} type cannot specify itself as an underlying type.")

            default_underlying_type = self.get_type(underlying_type_name)
            if not isinstance(default_underlying_type, IntegerType):
                raise RuntimeError(
                    f"{default_underlying_type.name} is not a numeric type, so it cannot be "
                    + "specified as an underlying type."
                )

            underlying_type = default_underlying_type

        protocol_values = protocol_enum.findall("value")

        values = []
        ordinals = set()
        names = set()

        for protocol_value in protocol_values:
            text = get_text(protocol_value)
            ordinal = try_parse_int(text)
            value_name = get_required_string_attribute(protocol_value, "name")
            python_name = value_name

            if python_name == 'None':
                python_name += '_'

            if ordinal is None:
                raise RuntimeError(f'{enum_name}.{value_name} has invalid ordinal value "{text}"')

            if ordinal not in ordinals:
                ordinals.add(ordinal)
            else:
                raise RuntimeError(
                    f'{enum_name}.{value_name} cannot redefine ordinal value {ordinal}.'
                )

            if python_name not in names:
                names.add(python_name)
            else:
                raise RuntimeError(f"{enum_name} enum cannot redefine value name {value_name}.")

            values.append(EnumValue(ordinal, value_name, python_name))

        return EnumType(enum_name, relative_path, underlying_type, values)

    def _create_struct_type(self, protocol_struct, relative_path):
        return StructType(
            get_required_string_attribute(protocol_struct, "name"),
            self._calculate_fixed_struct_size(protocol_struct),
            self._is_bounded(protocol_struct),
            relative_path,
        )

    def _calculate_fixed_struct_size(self, protocol_struct):
        size = 0

        for instruction in TypeFactory._flatten_instructions(protocol_struct):
            instruction_size = 0

            if instruction.tag == "field":
                instruction_size = self._calculate_fixed_struct_field_size(instruction)
            elif instruction.tag == "array":
                instruction_size = self._calculate_fixed_struct_array_size(instruction)
            elif instruction.tag == "dummy":
                instruction_size = self._calculate_fixed_struct_dummy_size(instruction)
            elif instruction.tag == "chunked":
                # Chunked reading is not allowed in fixed-size structs
                return None
            elif instruction.tag == "switch":
                # Switch sections are not allowed in fixed-sized structs
                return None

            if instruction_size is None:
                return None

            size += instruction_size

        return size

    def _calculate_fixed_struct_field_size(self, protocol_field):
        type_name = get_required_string_attribute(protocol_field, "type")
        type_length = TypeFactory._create_type_length_for_field(protocol_field)
        type_instance = self.get_type(type_name, type_length)
        field_size = type_instance.fixed_size

        if field_size is None:
            # All fields in a fixed-size struct must also be fixed-size
            return None

        if protocol_field.get("optional"):
            # Nothing can be optional in a fixed-size struct
            return None

        return field_size

    def _calculate_fixed_struct_array_size(self, protocol_array):
        length_string = protocol_array.get("length")
        length = try_parse_int(length_string)

        if length is None:
            # An array cannot be fixed-size unless a numeric length attribute is provided
            return None

        type_name = get_required_string_attribute(protocol_array, "type")
        type_instance = self.get_type(type_name)

        element_size = type_instance.fixed_size

        if element_size is None:
            # An array cannot be fixed-size unless its elements are also fixed-size
            # All arrays in a fixed-size struct must also be fixed-size
            return None

        if protocol_array.get("optional"):
            # Nothing can be optional in a fixed-size struct
            return None

        if protocol_array.get("delimited"):
            # It's possible to omit data or insert garbage data at the end of each chunk
            return None

        return length * element_size

    def _calculate_fixed_struct_dummy_size(self, protocol_dummy):
        type_name = get_required_string_attribute(protocol_dummy, "type")
        type_instance = self.get_type(type_name)
        dummy_size = type_instance.fixed_size

        if dummy_size is None:
            # All dummy fields in a fixed-size struct must also be fixed-size
            return None

        return dummy_size

    def _is_bounded(self, protocol_struct):
        result = True

        for instruction in TypeFactory._flatten_instructions(protocol_struct):
            if not result:
                result = instruction.tag == "break"
                continue

            if instruction.tag == "field":
                field_type = self.get_type(
                    get_required_string_attribute(instruction, "type"),
                    TypeFactory._create_type_length_for_field(instruction),
                )
                result = field_type.bounded
            elif instruction.tag == "array":
                element_type = self.get_type(get_required_string_attribute(instruction, "type"))
                length = get_string_attribute(instruction, "length")
                result = element_type.bounded and length is not None
            elif instruction.tag == "dummy":
                dummy_type = self.get_type(get_required_string_attribute(instruction, "type"))
                result = dummy_type.bounded

        return result

    @staticmethod
    def _flatten_instruction(instruction, result):
        result.append(instruction)

        if instruction.tag == "chunked":
            for chunked_instruction in get_instructions(instruction):
                TypeFactory._flatten_instruction(chunked_instruction, result)
        elif instruction.tag == "switch":
            protocol_cases = instruction.findall("case")
            for protocol_case in protocol_cases:
                for case_instruction in get_instructions(protocol_case):
                    TypeFactory._flatten_instruction(case_instruction, result)

    @staticmethod
    def _flatten_instructions(element, result=None):
        if result is None:
            result = []

        for instruction in get_instructions(element):
            TypeFactory._flatten_instruction(instruction, result)

        return result

    @staticmethod
    def _create_type_length_for_field(protocol_field):
        length_string = protocol_field.get("length")
        if length_string is not None:
            return Length.from_string(length_string)
        else:
            return Length.unspecified()

    @staticmethod
    def _create_type_with_specified_length(name, length):
        if name in ["string", "encoded_string"]:
            return StringType(name, length)
        else:
            raise RuntimeError(
                f'{name} type with length {length} is invalid. '
                + '(Only string types may specify a length)'
            )


class UnresolvedCustomType:
    def __init__(self, type_xml, relative_path):
        self._type_xml = type_xml
        self._relative_path = relative_path

    @property
    def type_xml(self):
        return self._type_xml

    @property
    def relative_path(self):
        return self._relative_path



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class InitReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with INIT_INIT packet
    """
    OutOfDate = 1
    Ok = 2
    Banned = 3
    """
    The official client won't display a message until the connection from the server is closed
    """
    WarpMap = 4
    FileEmf = 5
    FileEif = 6
    FileEnf = 7
    FileEsf = 8
    PlayersList = 9
    MapMutation = 10
    PlayersListFriends = 11
    FileEcf = 12



import os
from collections import namedtuple


from protocol_code_generator.util.name_utils import pascal_case_to_snake_case


class Import:
    def __init__(self, import_name, absolute_package_path):
        self._import_name = import_name
        self._absolute_package_path = absolute_package_path

    def relativize(self, package_path):
        from_package_path = self._absolute_package_path

        if self._absolute_package_path.startswith('eolib.'):
            absolute_parts = self._absolute_package_path.split('.')
            base_parts = package_path.split('.')
            for _ in range(min(len(absolute_parts), len(base_parts))):
                if absolute_parts[0] == base_parts[0]:
                    absolute_parts.pop(0)
                    base_parts.pop(0)
                else:
                    break
            from_package_path = ('.' * (len(base_parts) + 1)) + '.'.join(absolute_parts)

        return f"from {from_package_path} import {self._import_name}"


class CodeBlock:
    def __init__(self):
        self._imports = set()
        self._lines = [""]
        self._indentation = 0

    def add(self, code):
        parts = code.split("\n")
        for i in range(len(parts)):
            if len(parts[i]) > 0:
                line_index = len(self._lines) - 1
                if len(self._lines[line_index]) == 0:
                    self._lines[line_index] = " " * (self._indentation * 4)
                self._lines[line_index] += parts[i]
            if i != len(parts) - 1:
                self._lines.append("")
        return self

    def add_line(self, line=""):
        self.add(f"{line}\n")
        return self

    def add_code_block(self, block):
        for i in block._imports:
            self._imports.add(i)

        for i in range(len(block._lines)):
            if i == len(block._lines) - 1:
                self.add(block._lines[i])
            else:
                self.add_line(block._lines[i])

        return self

    def add_import(self, import_name, absolute_package_path):
        self._imports.add(Import(import_name, absolute_package_path))
        return self

    def add_import_by_type(self, custom_type):
        relative_path = custom_type.source_path.replace('/', '.').lstrip('.')
        if relative_path:
            relative_path += '.'
        path = (
            'eolib.protocol._generated.'
            + relative_path
            + pascal_case_to_snake_case(custom_type.name)
        )
        self.add_import(custom_type.name, path)
        return self

    def begin_control_flow(self, control_flow):
        self.add_line(f"{control_flow}:")
        self.indent()
        return self

    def next_control_flow(self, control_flow):
        self.unindent()
        return self.begin_control_flow(control_flow)

    def indent(self):
        self._indentation += 1
        return self

    def unindent(self):
        self._indentation -= 1
        return self

    @property
    def empty(self):
        return len(self._lines) == 1 and len(self._lines[0]) == 0

    def __bool__(self):
        return not self.empty

    def to_string(self, package_path):
        result = ""

        import_strings = {i.relativize(package_path) for i in self._imports}
        import_strings = sorted(import_strings, reverse=True)

        for import_ in import_strings:
            if import_.startswith("from __future__"):
                import_strings.remove(import_)
                import_strings.insert(0, import_)

        for import_ in import_strings:
            result += import_ + "\n"

        if not self.empty:
            if len(result) > 0:
                result += "\n"

            result += "\n".join(self._lines)

        return result



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .npc_map_info import NpcMapInfo
from .item_map_info import ItemMapInfo
from .character_map_info import CharacterMapInfo
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class NearbyInfo:
    """
    Information about nearby entities
    """
    _byte_size: int = 0
    _characters_count: int
    _characters: tuple[CharacterMapInfo, ...]
    _npcs: tuple[NpcMapInfo, ...]
    _items: tuple[ItemMapInfo, ...]

    def __init__(self, *, characters: Iterable[CharacterMapInfo], npcs: Iterable[NpcMapInfo], items: Iterable[ItemMapInfo]):
        """
        Create a new instance of NearbyInfo.

        Args:
            characters (Iterable[CharacterMapInfo]): (Length must be 252 or less.)
            npcs (Iterable[NpcMapInfo]): 
            items (Iterable[ItemMapInfo]): 
        """
        self._characters = tuple(characters)
        self._characters_count = len(self._characters)
        self._npcs = tuple(npcs)
        self._items = tuple(items)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def characters(self) -> tuple[CharacterMapInfo, ...]:
        return self._characters

    @property
    def npcs(self) -> tuple[NpcMapInfo, ...]:
        return self._npcs

    @property
    def items(self) -> tuple[ItemMapInfo, ...]:
        return self._items

    @staticmethod
    def serialize(writer: EoWriter, data: "NearbyInfo") -> None:
        """
        Serializes an instance of `NearbyInfo` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (NearbyInfo): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._characters_count is None:
                raise SerializationError("characters_count must be provided.")
            writer.add_char(data._characters_count)
            writer.string_sanitization_mode = True
            writer.add_byte(0xFF)
            if data._characters is None:
                raise SerializationError("characters must be provided.")
            if len(data._characters) > 252:
                raise SerializationError(f"Expected length of characters to be 252 or less, got {len(data._characters)}.")
            for i in range(data._characters_count):
                CharacterMapInfo.serialize(writer, data._characters[i])
                writer.add_byte(0xFF)
            if data._npcs is None:
                raise SerializationError("npcs must be provided.")
            for i in range(len(data._npcs)):
                NpcMapInfo.serialize(writer, data._npcs[i])
            writer.add_byte(0xFF)
            if data._items is None:
                raise SerializationError("items must be provided.")
            for i in range(len(data._items)):
                ItemMapInfo.serialize(writer, data._items[i])
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "NearbyInfo":
        """
        Deserializes an instance of `NearbyInfo` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            NearbyInfo: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            characters_count = reader.get_char()
            reader.chunked_reading_mode = True
            reader.next_chunk()
            characters = []
            for i in range(characters_count):
                characters.append(CharacterMapInfo.deserialize(reader))
                reader.next_chunk()
            npcs_length = int(reader.remaining / 6)
            npcs = []
            for i in range(npcs_length):
                npcs.append(NpcMapInfo.deserialize(reader))
            reader.next_chunk()
            items_length = int(reader.remaining / 9)
            items = []
            for i in range(items_length):
                items.append(ItemMapInfo.deserialize(reader))
            reader.chunked_reading_mode = False
            result = NearbyInfo(characters=characters, npcs=npcs, items=items)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"NearbyInfo(byte_size={repr(self._byte_size)}, characters={repr(self._characters)}, npcs={repr(self._npcs)}, items={repr(self._items)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..coords import Coords
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class MapLegacyDoorKey:
    """
    Legacy EMF entity used to specify a key on a door
    """
    _byte_size: int = 0
    _coords: Coords
    _key: int

    def __init__(self, *, coords: Coords, key: int):
        """
        Create a new instance of MapLegacyDoorKey.

        Args:
            coords (Coords): 
            key (int): (Value range is 0-64008.)
        """
        self._coords = coords
        self._key = key

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def coords(self) -> Coords:
        return self._coords

    @property
    def key(self) -> int:
        return self._key

    @staticmethod
    def serialize(writer: EoWriter, data: "MapLegacyDoorKey") -> None:
        """
        Serializes an instance of `MapLegacyDoorKey` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (MapLegacyDoorKey): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._coords is None:
                raise SerializationError("coords must be provided.")
            Coords.serialize(writer, data._coords)
            if data._key is None:
                raise SerializationError("key must be provided.")
            writer.add_short(data._key)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "MapLegacyDoorKey":
        """
        Deserializes an instance of `MapLegacyDoorKey` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            MapLegacyDoorKey: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            coords = Coords.deserialize(reader)
            key = reader.get_short()
            result = MapLegacyDoorKey(coords=coords, key=key)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"MapLegacyDoorKey(byte_size={repr(self._byte_size)}, coords={repr(self._coords)}, key={repr(self._key)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from typing import Union
from collections.abc import Iterable
from .welcome_code import WelcomeCode
from .server_settings import ServerSettings
from .nearby_info import NearbyInfo
from .login_message_code import LoginMessageCode
from .equipment_welcome import EquipmentWelcome
from .character_stats_welcome import CharacterStatsWelcome
from ..weight import Weight
from ..spell import Spell
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ..item import Item
from ...admin_level import AdminLevel
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class WelcomeReplyServerPacket(Packet):
    """
    Reply to selecting a character / entering game
    """
    _byte_size: int = 0
    _welcome_code: WelcomeCode
    _welcome_code_data: 'WelcomeReplyServerPacket.WelcomeCodeData'

    def __init__(self, *, welcome_code: WelcomeCode, welcome_code_data: 'WelcomeReplyServerPacket.WelcomeCodeData' = None):
        """
        Create a new instance of WelcomeReplyServerPacket.

        Args:
            welcome_code (WelcomeCode): 
            welcome_code_data (WelcomeReplyServerPacket.WelcomeCodeData): Data associated with the `welcome_code` field.
        """
        self._welcome_code = welcome_code
        self._welcome_code_data = welcome_code_data

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def welcome_code(self) -> WelcomeCode:
        return self._welcome_code

    @property
    def welcome_code_data(self) -> 'WelcomeReplyServerPacket.WelcomeCodeData':
        """
        WelcomeReplyServerPacket.WelcomeCodeData: Data associated with the `welcome_code` field.
        """
        return self._welcome_code_data

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Welcome

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Reply

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        WelcomeReplyServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "WelcomeReplyServerPacket") -> None:
        """
        Serializes an instance of `WelcomeReplyServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (WelcomeReplyServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._welcome_code is None:
                raise SerializationError("welcome_code must be provided.")
            writer.add_short(int(data._welcome_code))
            writer.string_sanitization_mode = True
            if data._welcome_code == WelcomeCode.SelectCharacter:
                if not isinstance(data._welcome_code_data, WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter):
                    raise SerializationError("Expected welcome_code_data to be type WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter for welcome_code " + WelcomeCode(data._welcome_code).name + ".")
                WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter.serialize(writer, data._welcome_code_data)
            elif data._welcome_code == WelcomeCode.EnterGame:
                if not isinstance(data._welcome_code_data, WelcomeReplyServerPacket.WelcomeCodeDataEnterGame):
                    raise SerializationError("Expected welcome_code_data to be type WelcomeReplyServerPacket.WelcomeCodeDataEnterGame for welcome_code " + WelcomeCode(data._welcome_code).name + ".")
                WelcomeReplyServerPacket.WelcomeCodeDataEnterGame.serialize(writer, data._welcome_code_data)
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "WelcomeReplyServerPacket":
        """
        Deserializes an instance of `WelcomeReplyServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            WelcomeReplyServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            welcome_code = WelcomeCode(reader.get_short())
            reader.chunked_reading_mode = True
            welcome_code_data: WelcomeReplyServerPacket.WelcomeCodeData = None
            if welcome_code == WelcomeCode.SelectCharacter:
                welcome_code_data = WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter.deserialize(reader)
            elif welcome_code == WelcomeCode.EnterGame:
                welcome_code_data = WelcomeReplyServerPacket.WelcomeCodeDataEnterGame.deserialize(reader)
            reader.chunked_reading_mode = False
            result = WelcomeReplyServerPacket(welcome_code=welcome_code, welcome_code_data=welcome_code_data)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"WelcomeReplyServerPacket(byte_size={repr(self._byte_size)}, welcome_code={repr(self._welcome_code)}, welcome_code_data={repr(self._welcome_code_data)})"

    WelcomeCodeData = Union['WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter', 'WelcomeReplyServerPacket.WelcomeCodeDataEnterGame', None]
    """
    Data associated with different values of the `welcome_code` field.
    """

    class WelcomeCodeDataSelectCharacter:
        """
        Data associated with welcome_code value WelcomeCode.SelectCharacter
        """
        _byte_size: int = 0
        _session_id: int
        _character_id: int
        _map_id: int
        _map_rid: tuple[int, ...]
        _map_file_size: int
        _eif_rid: tuple[int, ...]
        _eif_length: int
        _enf_rid: tuple[int, ...]
        _enf_length: int
        _esf_rid: tuple[int, ...]
        _esf_length: int
        _ecf_rid: tuple[int, ...]
        _ecf_length: int
        _name: str
        _title: str
        _guild_name: str
        _guild_rank_name: str
        _class_id: int
        _guild_tag: str
        _admin: AdminLevel
        _level: int
        _experience: int
        _usage: int
        _stats: CharacterStatsWelcome
        _equipment: EquipmentWelcome
        _guild_rank: int
        _settings: ServerSettings
        _login_message_code: LoginMessageCode

        def __init__(self, *, session_id: int, character_id: int, map_id: int, map_rid: Iterable[int], map_file_size: int, eif_rid: Iterable[int], eif_length: int, enf_rid: Iterable[int], enf_length: int, esf_rid: Iterable[int], esf_length: int, ecf_rid: Iterable[int], ecf_length: int, name: str, title: str, guild_name: str, guild_rank_name: str, class_id: int, guild_tag: str, admin: AdminLevel, level: int, experience: int, usage: int, stats: CharacterStatsWelcome, equipment: EquipmentWelcome, guild_rank: int, settings: ServerSettings, login_message_code: LoginMessageCode):
            """
            Create a new instance of WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter.

            Args:
                session_id (int): (Value range is 0-64008.)
                character_id (int): (Value range is 0-4097152080.)
                map_id (int): (Value range is 0-64008.)
                map_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                map_file_size (int): (Value range is 0-16194276.)
                eif_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                eif_length (int): (Value range is 0-64008.)
                enf_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                enf_length (int): (Value range is 0-64008.)
                esf_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                esf_length (int): (Value range is 0-64008.)
                ecf_rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
                ecf_length (int): (Value range is 0-64008.)
                name (str): 
                title (str): 
                guild_name (str): 
                guild_rank_name (str): 
                class_id (int): (Value range is 0-252.)
                guild_tag (str): (Length must be `3`.)
                admin (AdminLevel): 
                level (int): (Value range is 0-252.)
                experience (int): (Value range is 0-4097152080.)
                usage (int): (Value range is 0-4097152080.)
                stats (CharacterStatsWelcome): 
                equipment (EquipmentWelcome): 
                guild_rank (int): (Value range is 0-252.)
                settings (ServerSettings): 
                login_message_code (LoginMessageCode): 
            """
            self._session_id = session_id
            self._character_id = character_id
            self._map_id = map_id
            self._map_rid = tuple(map_rid)
            self._map_file_size = map_file_size
            self._eif_rid = tuple(eif_rid)
            self._eif_length = eif_length
            self._enf_rid = tuple(enf_rid)
            self._enf_length = enf_length
            self._esf_rid = tuple(esf_rid)
            self._esf_length = esf_length
            self._ecf_rid = tuple(ecf_rid)
            self._ecf_length = ecf_length
            self._name = name
            self._title = title
            self._guild_name = guild_name
            self._guild_rank_name = guild_rank_name
            self._class_id = class_id
            self._guild_tag = guild_tag
            self._admin = admin
            self._level = level
            self._experience = experience
            self._usage = usage
            self._stats = stats
            self._equipment = equipment
            self._guild_rank = guild_rank
            self._settings = settings
            self._login_message_code = login_message_code

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def session_id(self) -> int:
            return self._session_id

        @property
        def character_id(self) -> int:
            return self._character_id

        @property
        def map_id(self) -> int:
            return self._map_id

        @property
        def map_rid(self) -> tuple[int, ...]:
            return self._map_rid

        @property
        def map_file_size(self) -> int:
            return self._map_file_size

        @property
        def eif_rid(self) -> tuple[int, ...]:
            return self._eif_rid

        @property
        def eif_length(self) -> int:
            return self._eif_length

        @property
        def enf_rid(self) -> tuple[int, ...]:
            return self._enf_rid

        @property
        def enf_length(self) -> int:
            return self._enf_length

        @property
        def esf_rid(self) -> tuple[int, ...]:
            return self._esf_rid

        @property
        def esf_length(self) -> int:
            return self._esf_length

        @property
        def ecf_rid(self) -> tuple[int, ...]:
            return self._ecf_rid

        @property
        def ecf_length(self) -> int:
            return self._ecf_length

        @property
        def name(self) -> str:
            return self._name

        @property
        def title(self) -> str:
            return self._title

        @property
        def guild_name(self) -> str:
            return self._guild_name

        @property
        def guild_rank_name(self) -> str:
            return self._guild_rank_name

        @property
        def class_id(self) -> int:
            return self._class_id

        @property
        def guild_tag(self) -> str:
            return self._guild_tag

        @property
        def admin(self) -> AdminLevel:
            return self._admin

        @property
        def level(self) -> int:
            return self._level

        @property
        def experience(self) -> int:
            return self._experience

        @property
        def usage(self) -> int:
            return self._usage

        @property
        def stats(self) -> CharacterStatsWelcome:
            return self._stats

        @property
        def equipment(self) -> EquipmentWelcome:
            return self._equipment

        @property
        def guild_rank(self) -> int:
            return self._guild_rank

        @property
        def settings(self) -> ServerSettings:
            return self._settings

        @property
        def login_message_code(self) -> LoginMessageCode:
            return self._login_message_code

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter") -> None:
            """
            Serializes an instance of `WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                if data._session_id is None:
                    raise SerializationError("session_id must be provided.")
                writer.add_short(data._session_id)
                if data._character_id is None:
                    raise SerializationError("character_id must be provided.")
                writer.add_int(data._character_id)
                if data._map_id is None:
                    raise SerializationError("map_id must be provided.")
                writer.add_short(data._map_id)
                if data._map_rid is None:
                    raise SerializationError("map_rid must be provided.")
                if len(data._map_rid) != 2:
                    raise SerializationError(f"Expected length of map_rid to be exactly 2, got {len(data._map_rid)}.")
                for i in range(2):
                    writer.add_short(data._map_rid[i])
                if data._map_file_size is None:
                    raise SerializationError("map_file_size must be provided.")
                writer.add_three(data._map_file_size)
                if data._eif_rid is None:
                    raise SerializationError("eif_rid must be provided.")
                if len(data._eif_rid) != 2:
                    raise SerializationError(f"Expected length of eif_rid to be exactly 2, got {len(data._eif_rid)}.")
                for i in range(2):
                    writer.add_short(data._eif_rid[i])
                if data._eif_length is None:
                    raise SerializationError("eif_length must be provided.")
                writer.add_short(data._eif_length)
                if data._enf_rid is None:
                    raise SerializationError("enf_rid must be provided.")
                if len(data._enf_rid) != 2:
                    raise SerializationError(f"Expected length of enf_rid to be exactly 2, got {len(data._enf_rid)}.")
                for i in range(2):
                    writer.add_short(data._enf_rid[i])
                if data._enf_length is None:
                    raise SerializationError("enf_length must be provided.")
                writer.add_short(data._enf_length)
                if data._esf_rid is None:
                    raise SerializationError("esf_rid must be provided.")
                if len(data._esf_rid) != 2:
                    raise SerializationError(f"Expected length of esf_rid to be exactly 2, got {len(data._esf_rid)}.")
                for i in range(2):
                    writer.add_short(data._esf_rid[i])
                if data._esf_length is None:
                    raise SerializationError("esf_length must be provided.")
                writer.add_short(data._esf_length)
                if data._ecf_rid is None:
                    raise SerializationError("ecf_rid must be provided.")
                if len(data._ecf_rid) != 2:
                    raise SerializationError(f"Expected length of ecf_rid to be exactly 2, got {len(data._ecf_rid)}.")
                for i in range(2):
                    writer.add_short(data._ecf_rid[i])
                if data._ecf_length is None:
                    raise SerializationError("ecf_length must be provided.")
                writer.add_short(data._ecf_length)
                if data._name is None:
                    raise SerializationError("name must be provided.")
                writer.add_string(data._name)
                writer.add_byte(0xFF)
                if data._title is None:
                    raise SerializationError("title must be provided.")
                writer.add_string(data._title)
                writer.add_byte(0xFF)
                if data._guild_name is None:
                    raise SerializationError("guild_name must be provided.")
                writer.add_string(data._guild_name)
                writer.add_byte(0xFF)
                if data._guild_rank_name is None:
                    raise SerializationError("guild_rank_name must be provided.")
                writer.add_string(data._guild_rank_name)
                writer.add_byte(0xFF)
                if data._class_id is None:
                    raise SerializationError("class_id must be provided.")
                writer.add_char(data._class_id)
                if data._guild_tag is None:
                    raise SerializationError("guild_tag must be provided.")
                if len(data._guild_tag) != 3:
                    raise SerializationError(f"Expected length of guild_tag to be exactly 3, got {len(data._guild_tag)}.")
                writer.add_fixed_string(data._guild_tag, 3, False)
                if data._admin is None:
                    raise SerializationError("admin must be provided.")
                writer.add_char(int(data._admin))
                if data._level is None:
                    raise SerializationError("level must be provided.")
                writer.add_char(data._level)
                if data._experience is None:
                    raise SerializationError("experience must be provided.")
                writer.add_int(data._experience)
                if data._usage is None:
                    raise SerializationError("usage must be provided.")
                writer.add_int(data._usage)
                if data._stats is None:
                    raise SerializationError("stats must be provided.")
                CharacterStatsWelcome.serialize(writer, data._stats)
                if data._equipment is None:
                    raise SerializationError("equipment must be provided.")
                EquipmentWelcome.serialize(writer, data._equipment)
                if data._guild_rank is None:
                    raise SerializationError("guild_rank must be provided.")
                writer.add_char(data._guild_rank)
                if data._settings is None:
                    raise SerializationError("settings must be provided.")
                ServerSettings.serialize(writer, data._settings)
                if data._login_message_code is None:
                    raise SerializationError("login_message_code must be provided.")
                writer.add_char(int(data._login_message_code))
                writer.add_byte(0xFF)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter":
            """
            Deserializes an instance of `WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                session_id = reader.get_short()
                character_id = reader.get_int()
                map_id = reader.get_short()
                map_rid = []
                for i in range(2):
                    map_rid.append(reader.get_short())
                map_file_size = reader.get_three()
                eif_rid = []
                for i in range(2):
                    eif_rid.append(reader.get_short())
                eif_length = reader.get_short()
                enf_rid = []
                for i in range(2):
                    enf_rid.append(reader.get_short())
                enf_length = reader.get_short()
                esf_rid = []
                for i in range(2):
                    esf_rid.append(reader.get_short())
                esf_length = reader.get_short()
                ecf_rid = []
                for i in range(2):
                    ecf_rid.append(reader.get_short())
                ecf_length = reader.get_short()
                name = reader.get_string()
                reader.next_chunk()
                title = reader.get_string()
                reader.next_chunk()
                guild_name = reader.get_string()
                reader.next_chunk()
                guild_rank_name = reader.get_string()
                reader.next_chunk()
                class_id = reader.get_char()
                guild_tag = reader.get_fixed_string(3, False)
                admin = AdminLevel(reader.get_char())
                level = reader.get_char()
                experience = reader.get_int()
                usage = reader.get_int()
                stats = CharacterStatsWelcome.deserialize(reader)
                equipment = EquipmentWelcome.deserialize(reader)
                guild_rank = reader.get_char()
                settings = ServerSettings.deserialize(reader)
                login_message_code = LoginMessageCode(reader.get_char())
                reader.next_chunk()
                result = WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter(session_id=session_id, character_id=character_id, map_id=map_id, map_rid=map_rid, map_file_size=map_file_size, eif_rid=eif_rid, eif_length=eif_length, enf_rid=enf_rid, enf_length=enf_length, esf_rid=esf_rid, esf_length=esf_length, ecf_rid=ecf_rid, ecf_length=ecf_length, name=name, title=title, guild_name=guild_name, guild_rank_name=guild_rank_name, class_id=class_id, guild_tag=guild_tag, admin=admin, level=level, experience=experience, usage=usage, stats=stats, equipment=equipment, guild_rank=guild_rank, settings=settings, login_message_code=login_message_code)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeReplyServerPacket.WelcomeCodeDataSelectCharacter(byte_size={repr(self._byte_size)}, session_id={repr(self._session_id)}, character_id={repr(self._character_id)}, map_id={repr(self._map_id)}, map_rid={repr(self._map_rid)}, map_file_size={repr(self._map_file_size)}, eif_rid={repr(self._eif_rid)}, eif_length={repr(self._eif_length)}, enf_rid={repr(self._enf_rid)}, enf_length={repr(self._enf_length)}, esf_rid={repr(self._esf_rid)}, esf_length={repr(self._esf_length)}, ecf_rid={repr(self._ecf_rid)}, ecf_length={repr(self._ecf_length)}, name={repr(self._name)}, title={repr(self._title)}, guild_name={repr(self._guild_name)}, guild_rank_name={repr(self._guild_rank_name)}, class_id={repr(self._class_id)}, guild_tag={repr(self._guild_tag)}, admin={repr(self._admin)}, level={repr(self._level)}, experience={repr(self._experience)}, usage={repr(self._usage)}, stats={repr(self._stats)}, equipment={repr(self._equipment)}, guild_rank={repr(self._guild_rank)}, settings={repr(self._settings)}, login_message_code={repr(self._login_message_code)})"

    class WelcomeCodeDataEnterGame:
        """
        Data associated with welcome_code value WelcomeCode.EnterGame
        """
        _byte_size: int = 0
        _news: tuple[str, ...]
        _weight: Weight
        _items: tuple[Item, ...]
        _spells: tuple[Spell, ...]
        _nearby: NearbyInfo

        def __init__(self, *, news: Iterable[str], weight: Weight, items: Iterable[Item], spells: Iterable[Spell], nearby: NearbyInfo):
            """
            Create a new instance of WelcomeReplyServerPacket.WelcomeCodeDataEnterGame.

            Args:
                news (Iterable[str]): (Length must be `9`.)
                weight (Weight): 
                items (Iterable[Item]): 
                spells (Iterable[Spell]): 
                nearby (NearbyInfo): 
            """
            self._news = tuple(news)
            self._weight = weight
            self._items = tuple(items)
            self._spells = tuple(spells)
            self._nearby = nearby

        @property
        def byte_size(self) -> int:
            """
            Returns the size of the data that this was deserialized from.

            Returns:
                int: The size of the data that this was deserialized from.
            """
            return self._byte_size

        @property
        def news(self) -> tuple[str, ...]:
            return self._news

        @property
        def weight(self) -> Weight:
            return self._weight

        @property
        def items(self) -> tuple[Item, ...]:
            return self._items

        @property
        def spells(self) -> tuple[Spell, ...]:
            return self._spells

        @property
        def nearby(self) -> NearbyInfo:
            return self._nearby

        @staticmethod
        def serialize(writer: EoWriter, data: "WelcomeReplyServerPacket.WelcomeCodeDataEnterGame") -> None:
            """
            Serializes an instance of `WelcomeReplyServerPacket.WelcomeCodeDataEnterGame` to the provided `EoWriter`.

            Args:
                writer (EoWriter): The writer that the data will be serialized to.
                data (WelcomeReplyServerPacket.WelcomeCodeDataEnterGame): The data to serialize.
            """
            old_string_sanitization_mode: bool = writer.string_sanitization_mode
            try:
                writer.add_byte(0xFF)
                if data._news is None:
                    raise SerializationError("news must be provided.")
                if len(data._news) != 9:
                    raise SerializationError(f"Expected length of news to be exactly 9, got {len(data._news)}.")
                for i in range(9):
                    writer.add_string(data._news[i])
                    writer.add_byte(0xFF)
                if data._weight is None:
                    raise SerializationError("weight must be provided.")
                Weight.serialize(writer, data._weight)
                if data._items is None:
                    raise SerializationError("items must be provided.")
                for i in range(len(data._items)):
                    Item.serialize(writer, data._items[i])
                writer.add_byte(0xFF)
                if data._spells is None:
                    raise SerializationError("spells must be provided.")
                for i in range(len(data._spells)):
                    Spell.serialize(writer, data._spells[i])
                writer.add_byte(0xFF)
                if data._nearby is None:
                    raise SerializationError("nearby must be provided.")
                NearbyInfo.serialize(writer, data._nearby)
            finally:
                writer.string_sanitization_mode = old_string_sanitization_mode

        @staticmethod
        def deserialize(reader: EoReader) -> "WelcomeReplyServerPacket.WelcomeCodeDataEnterGame":
            """
            Deserializes an instance of `WelcomeReplyServerPacket.WelcomeCodeDataEnterGame` from the provided `EoReader`.

            Args:
                reader (EoReader): The writer that the data will be serialized to.

            Returns:
                WelcomeReplyServerPacket.WelcomeCodeDataEnterGame: The data to serialize.
            """
            old_chunked_reading_mode: bool = reader.chunked_reading_mode
            try:
                reader_start_position: int = reader.position
                reader.next_chunk()
                news = []
                for i in range(9):
                    news.append(reader.get_string())
                    reader.next_chunk()
                weight = Weight.deserialize(reader)
                items_length = int(reader.remaining / 6)
                items = []
                for i in range(items_length):
                    items.append(Item.deserialize(reader))
                reader.next_chunk()
                spells_length = int(reader.remaining / 4)
                spells = []
                for i in range(spells_length):
                    spells.append(Spell.deserialize(reader))
                reader.next_chunk()
                nearby = NearbyInfo.deserialize(reader)
                result = WelcomeReplyServerPacket.WelcomeCodeDataEnterGame(news=news, weight=weight, items=items, spells=spells, nearby=nearby)
                result._byte_size = reader.position - reader_start_position
                return result
            finally:
                reader.chunked_reading_mode = old_chunked_reading_mode

        def __repr__(self):
            return f"WelcomeReplyServerPacket.WelcomeCodeDataEnterGame(byte_size={repr(self._byte_size)}, news={repr(self._news)}, weight={repr(self._weight)}, items={repr(self._items)}, spells={repr(self._spells)}, nearby={repr(self._nearby)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class PartyReplyCode(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with PARTY_REPLY packet.
    Indicates why an invite or join request failed.
    """
    AlreadyInAnotherParty = 0
    AlreadyInYourParty = 1
    PartyIsFull = 2



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class MapFile:
    """
    A map file (EMF)
    """
    _byte_size: int = 0
    _content: bytes

    def __init__(self, *, content: bytes):
        """
        Create a new instance of MapFile.

        Args:
            content (bytes): 
        """
        self._content = content

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def content(self) -> bytes:
        return self._content

    @staticmethod
    def serialize(writer: EoWriter, data: "MapFile") -> None:
        """
        Serializes an instance of `MapFile` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (MapFile): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._content is None:
                raise SerializationError("content must be provided.")
            writer.add_bytes(data._content)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "MapFile":
        """
        Deserializes an instance of `MapFile` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            MapFile: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            content = reader.get_bytes(reader.remaining)
            result = MapFile(content=content)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"MapFile(byte_size={repr(self._byte_size)}, content={repr(self._content)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .shop_record import ShopRecord
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ShopFile:
    """
    Endless Shop File
    """
    _byte_size: int = 0
    _shops: tuple[ShopRecord, ...]

    def __init__(self, *, shops: Iterable[ShopRecord]):
        """
        Create a new instance of ShopFile.

        Args:
            shops (Iterable[ShopRecord]): 
        """
        self._shops = tuple(shops)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def shops(self) -> tuple[ShopRecord, ...]:
        return self._shops

    @staticmethod
    def serialize(writer: EoWriter, data: "ShopFile") -> None:
        """
        Serializes an instance of `ShopFile` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ShopFile): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.add_fixed_string("ESF", 3, False)
            if data._shops is None:
                raise SerializationError("shops must be provided.")
            for i in range(len(data._shops)):
                ShopRecord.serialize(writer, data._shops[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ShopFile":
        """
        Deserializes an instance of `ShopFile` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ShopFile: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.get_fixed_string(3, False)
            shops = []
            while reader.remaining > 0:
                shops.append(ShopRecord.deserialize(reader))
            result = ShopFile(shops=shops)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ShopFile(byte_size={repr(self._byte_size)}, shops={repr(self._shops)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .drop_npc_record import DropNpcRecord
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class DropFile:
    """
    Endless Drop File
    """
    _byte_size: int = 0
    _npcs: tuple[DropNpcRecord, ...]

    def __init__(self, *, npcs: Iterable[DropNpcRecord]):
        """
        Create a new instance of DropFile.

        Args:
            npcs (Iterable[DropNpcRecord]): 
        """
        self._npcs = tuple(npcs)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def npcs(self) -> tuple[DropNpcRecord, ...]:
        return self._npcs

    @staticmethod
    def serialize(writer: EoWriter, data: "DropFile") -> None:
        """
        Serializes an instance of `DropFile` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (DropFile): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.add_fixed_string("EDF", 3, False)
            if data._npcs is None:
                raise SerializationError("npcs must be provided.")
            for i in range(len(data._npcs)):
                DropNpcRecord.serialize(writer, data._npcs[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "DropFile":
        """
        Deserializes an instance of `DropFile` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            DropFile: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.get_fixed_string(3, False)
            npcs = []
            while reader.remaining > 0:
                npcs.append(DropNpcRecord.deserialize(reader))
            result = DropFile(npcs=npcs)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"DropFile(byte_size={repr(self._byte_size)}, npcs={repr(self._npcs)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from .map_warp_row import MapWarpRow
from .map_type import MapType
from .map_timed_effect import MapTimedEffect
from .map_tile_spec_row import MapTileSpecRow
from .map_sign import MapSign
from .map_npc import MapNpc
from .map_music_control import MapMusicControl
from .map_legacy_door_key import MapLegacyDoorKey
from .map_item import MapItem
from .map_graphic_layer import MapGraphicLayer
from ...serialization_error import SerializationError
from ....data.eo_writer import EoWriter
from ....data.eo_reader import EoReader

class Emf:
    """
    Endless Map File
    """
    _byte_size: int = 0
    _rid: tuple[int, ...]
    _name: str
    _type: MapType
    _timed_effect: MapTimedEffect
    _music_id: int
    _music_control: MapMusicControl
    _ambient_sound_id: int
    _width: int
    _height: int
    _fill_tile: int
    _map_available: bool
    _can_scroll: bool
    _relog_x: int
    _relog_y: int
    _npcs_count: int
    _npcs: tuple[MapNpc, ...]
    _legacy_door_keys_count: int
    _legacy_door_keys: tuple[MapLegacyDoorKey, ...]
    _items_count: int
    _items: tuple[MapItem, ...]
    _tile_spec_rows_count: int
    _tile_spec_rows: tuple[MapTileSpecRow, ...]
    _warp_rows_count: int
    _warp_rows: tuple[MapWarpRow, ...]
    _graphic_layers: tuple[MapGraphicLayer, ...]
    _signs_count: int
    _signs: tuple[MapSign, ...]

    def __init__(self, *, rid: Iterable[int], name: str, type: MapType, timed_effect: MapTimedEffect, music_id: int, music_control: MapMusicControl, ambient_sound_id: int, width: int, height: int, fill_tile: int, map_available: bool, can_scroll: bool, relog_x: int, relog_y: int, npcs: Iterable[MapNpc], legacy_door_keys: Iterable[MapLegacyDoorKey], items: Iterable[MapItem], tile_spec_rows: Iterable[MapTileSpecRow], warp_rows: Iterable[MapWarpRow], graphic_layers: Iterable[MapGraphicLayer], signs: Iterable[MapSign]):
        """
        Create a new instance of Emf.

        Args:
            rid (Iterable[int]): (Length must be `2`.) (Element value range is 0-64008.)
            name (str): (Length must be `24` or less.)
            type (MapType): 
            timed_effect (MapTimedEffect): 
            music_id (int): (Value range is 0-252.)
            music_control (MapMusicControl): 
            ambient_sound_id (int): (Value range is 0-64008.)
            width (int): (Value range is 0-252.)
            height (int): (Value range is 0-252.)
            fill_tile (int): (Value range is 0-64008.)
            map_available (bool): 
            can_scroll (bool): 
            relog_x (int): (Value range is 0-252.)
            relog_y (int): (Value range is 0-252.)
            npcs (Iterable[MapNpc]): (Length must be 252 or less.)
            legacy_door_keys (Iterable[MapLegacyDoorKey]): (Length must be 252 or less.)
            items (Iterable[MapItem]): (Length must be 252 or less.)
            tile_spec_rows (Iterable[MapTileSpecRow]): (Length must be 252 or less.)
            warp_rows (Iterable[MapWarpRow]): (Length must be 252 or less.)
            graphic_layers (Iterable[MapGraphicLayer]): The 9 layers of map graphics. Order is [Ground, Object, Overlay, Down Wall, Right Wall, Roof, Top, Shadow, Overlay2] (Length must be `9`.)
            signs (Iterable[MapSign]): (Length must be 252 or less.)
        """
        self._rid = tuple(rid)
        self._name = name
        self._type = type
        self._timed_effect = timed_effect
        self._music_id = music_id
        self._music_control = music_control
        self._ambient_sound_id = ambient_sound_id
        self._width = width
        self._height = height
        self._fill_tile = fill_tile
        self._map_available = map_available
        self._can_scroll = can_scroll
        self._relog_x = relog_x
        self._relog_y = relog_y
        self._npcs = tuple(npcs)
        self._npcs_count = len(self._npcs)
        self._legacy_door_keys = tuple(legacy_door_keys)
        self._legacy_door_keys_count = len(self._legacy_door_keys)
        self._items = tuple(items)
        self._items_count = len(self._items)
        self._tile_spec_rows = tuple(tile_spec_rows)
        self._tile_spec_rows_count = len(self._tile_spec_rows)
        self._warp_rows = tuple(warp_rows)
        self._warp_rows_count = len(self._warp_rows)
        self._graphic_layers = tuple(graphic_layers)
        self._signs = tuple(signs)
        self._signs_count = len(self._signs)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def rid(self) -> tuple[int, ...]:
        return self._rid

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> MapType:
        return self._type

    @property
    def timed_effect(self) -> MapTimedEffect:
        return self._timed_effect

    @property
    def music_id(self) -> int:
        return self._music_id

    @property
    def music_control(self) -> MapMusicControl:
        return self._music_control

    @property
    def ambient_sound_id(self) -> int:
        return self._ambient_sound_id

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def fill_tile(self) -> int:
        return self._fill_tile

    @property
    def map_available(self) -> bool:
        return self._map_available

    @property
    def can_scroll(self) -> bool:
        return self._can_scroll

    @property
    def relog_x(self) -> int:
        return self._relog_x

    @property
    def relog_y(self) -> int:
        return self._relog_y

    @property
    def npcs(self) -> tuple[MapNpc, ...]:
        return self._npcs

    @property
    def legacy_door_keys(self) -> tuple[MapLegacyDoorKey, ...]:
        return self._legacy_door_keys

    @property
    def items(self) -> tuple[MapItem, ...]:
        return self._items

    @property
    def tile_spec_rows(self) -> tuple[MapTileSpecRow, ...]:
        return self._tile_spec_rows

    @property
    def warp_rows(self) -> tuple[MapWarpRow, ...]:
        return self._warp_rows

    @property
    def graphic_layers(self) -> tuple[MapGraphicLayer, ...]:
        """
        The 9 layers of map graphics.
        Order is [Ground, Object, Overlay, Down Wall, Right Wall, Roof, Top, Shadow, Overlay2]
        """
        return self._graphic_layers

    @property
    def signs(self) -> tuple[MapSign, ...]:
        return self._signs

    @staticmethod
    def serialize(writer: EoWriter, data: "Emf") -> None:
        """
        Serializes an instance of `Emf` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (Emf): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.add_fixed_string("EMF", 3, False)
            if data._rid is None:
                raise SerializationError("rid must be provided.")
            if len(data._rid) != 2:
                raise SerializationError(f"Expected length of rid to be exactly 2, got {len(data._rid)}.")
            for i in range(2):
                writer.add_short(data._rid[i])
            if data._name is None:
                raise SerializationError("name must be provided.")
            if len(data._name) > 24:
                raise SerializationError(f"Expected length of name to be 24 or less, got {len(data._name)}.")
            writer.add_fixed_encoded_string(data._name, 24, True)
            if data._type is None:
                raise SerializationError("type must be provided.")
            writer.add_char(int(data._type))
            if data._timed_effect is None:
                raise SerializationError("timed_effect must be provided.")
            writer.add_char(int(data._timed_effect))
            if data._music_id is None:
                raise SerializationError("music_id must be provided.")
            writer.add_char(data._music_id)
            if data._music_control is None:
                raise SerializationError("music_control must be provided.")
            writer.add_char(int(data._music_control))
            if data._ambient_sound_id is None:
                raise SerializationError("ambient_sound_id must be provided.")
            writer.add_short(data._ambient_sound_id)
            if data._width is None:
                raise SerializationError("width must be provided.")
            writer.add_char(data._width)
            if data._height is None:
                raise SerializationError("height must be provided.")
            writer.add_char(data._height)
            if data._fill_tile is None:
                raise SerializationError("fill_tile must be provided.")
            writer.add_short(data._fill_tile)
            if data._map_available is None:
                raise SerializationError("map_available must be provided.")
            writer.add_char(1 if data._map_available else 0)
            if data._can_scroll is None:
                raise SerializationError("can_scroll must be provided.")
            writer.add_char(1 if data._can_scroll else 0)
            if data._relog_x is None:
                raise SerializationError("relog_x must be provided.")
            writer.add_char(data._relog_x)
            if data._relog_y is None:
                raise SerializationError("relog_y must be provided.")
            writer.add_char(data._relog_y)
            writer.add_char(0)
            if data._npcs_count is None:
                raise SerializationError("npcs_count must be provided.")
            writer.add_char(data._npcs_count)
            if data._npcs is None:
                raise SerializationError("npcs must be provided.")
            if len(data._npcs) > 252:
                raise SerializationError(f"Expected length of npcs to be 252 or less, got {len(data._npcs)}.")
            for i in range(data._npcs_count):
                MapNpc.serialize(writer, data._npcs[i])
            if data._legacy_door_keys_count is None:
                raise SerializationError("legacy_door_keys_count must be provided.")
            writer.add_char(data._legacy_door_keys_count)
            if data._legacy_door_keys is None:
                raise SerializationError("legacy_door_keys must be provided.")
            if len(data._legacy_door_keys) > 252:
                raise SerializationError(f"Expected length of legacy_door_keys to be 252 or less, got {len(data._legacy_door_keys)}.")
            for i in range(data._legacy_door_keys_count):
                MapLegacyDoorKey.serialize(writer, data._legacy_door_keys[i])
            if data._items_count is None:
                raise SerializationError("items_count must be provided.")
            writer.add_char(data._items_count)
            if data._items is None:
                raise SerializationError("items must be provided.")
            if len(data._items) > 252:
                raise SerializationError(f"Expected length of items to be 252 or less, got {len(data._items)}.")
            for i in range(data._items_count):
                MapItem.serialize(writer, data._items[i])
            if data._tile_spec_rows_count is None:
                raise SerializationError("tile_spec_rows_count must be provided.")
            writer.add_char(data._tile_spec_rows_count)
            if data._tile_spec_rows is None:
                raise SerializationError("tile_spec_rows must be provided.")
            if len(data._tile_spec_rows) > 252:
                raise SerializationError(f"Expected length of tile_spec_rows to be 252 or less, got {len(data._tile_spec_rows)}.")
            for i in range(data._tile_spec_rows_count):
                MapTileSpecRow.serialize(writer, data._tile_spec_rows[i])
            if data._warp_rows_count is None:
                raise SerializationError("warp_rows_count must be provided.")
            writer.add_char(data._warp_rows_count)
            if data._warp_rows is None:
                raise SerializationError("warp_rows must be provided.")
            if len(data._warp_rows) > 252:
                raise SerializationError(f"Expected length of warp_rows to be 252 or less, got {len(data._warp_rows)}.")
            for i in range(data._warp_rows_count):
                MapWarpRow.serialize(writer, data._warp_rows[i])
            if data._graphic_layers is None:
                raise SerializationError("graphic_layers must be provided.")
            if len(data._graphic_layers) != 9:
                raise SerializationError(f"Expected length of graphic_layers to be exactly 9, got {len(data._graphic_layers)}.")
            for i in range(9):
                MapGraphicLayer.serialize(writer, data._graphic_layers[i])
            if data._signs_count is None:
                raise SerializationError("signs_count must be provided.")
            writer.add_char(data._signs_count)
            if data._signs is None:
                raise SerializationError("signs must be provided.")
            if len(data._signs) > 252:
                raise SerializationError(f"Expected length of signs to be 252 or less, got {len(data._signs)}.")
            for i in range(data._signs_count):
                MapSign.serialize(writer, data._signs[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "Emf":
        """
        Deserializes an instance of `Emf` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            Emf: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.get_fixed_string(3, False)
            rid = []
            for i in range(2):
                rid.append(reader.get_short())
            name = reader.get_fixed_encoded_string(24, True)
            type = MapType(reader.get_char())
            timed_effect = MapTimedEffect(reader.get_char())
            music_id = reader.get_char()
            music_control = MapMusicControl(reader.get_char())
            ambient_sound_id = reader.get_short()
            width = reader.get_char()
            height = reader.get_char()
            fill_tile = reader.get_short()
            map_available = reader.get_char() != 0
            can_scroll = reader.get_char() != 0
            relog_x = reader.get_char()
            relog_y = reader.get_char()
            reader.get_char()
            npcs_count = reader.get_char()
            npcs = []
            for i in range(npcs_count):
                npcs.append(MapNpc.deserialize(reader))
            legacy_door_keys_count = reader.get_char()
            legacy_door_keys = []
            for i in range(legacy_door_keys_count):
                legacy_door_keys.append(MapLegacyDoorKey.deserialize(reader))
            items_count = reader.get_char()
            items = []
            for i in range(items_count):
                items.append(MapItem.deserialize(reader))
            tile_spec_rows_count = reader.get_char()
            tile_spec_rows = []
            for i in range(tile_spec_rows_count):
                tile_spec_rows.append(MapTileSpecRow.deserialize(reader))
            warp_rows_count = reader.get_char()
            warp_rows = []
            for i in range(warp_rows_count):
                warp_rows.append(MapWarpRow.deserialize(reader))
            graphic_layers = []
            for i in range(9):
                graphic_layers.append(MapGraphicLayer.deserialize(reader))
            signs_count = reader.get_char()
            signs = []
            for i in range(signs_count):
                signs.append(MapSign.deserialize(reader))
            result = Emf(rid=rid, name=name, type=type, timed_effect=timed_effect, music_id=music_id, music_control=music_control, ambient_sound_id=ambient_sound_id, width=width, height=height, fill_tile=fill_tile, map_available=map_available, can_scroll=can_scroll, relog_x=relog_x, relog_y=relog_y, npcs=npcs, legacy_door_keys=legacy_door_keys, items=items, tile_spec_rows=tile_spec_rows, warp_rows=warp_rows, graphic_layers=graphic_layers, signs=signs)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"Emf(byte_size={repr(self._byte_size)}, rid={repr(self._rid)}, name={repr(self._name)}, type={repr(self._type)}, timed_effect={repr(self._timed_effect)}, music_id={repr(self._music_id)}, music_control={repr(self._music_control)}, ambient_sound_id={repr(self._ambient_sound_id)}, width={repr(self._width)}, height={repr(self._height)}, fill_tile={repr(self._fill_tile)}, map_available={repr(self._map_available)}, can_scroll={repr(self._can_scroll)}, relog_x={repr(self._relog_x)}, relog_y={repr(self._relog_y)}, npcs={repr(self._npcs)}, legacy_door_keys={repr(self._legacy_door_keys)}, items={repr(self._items)}, tile_spec_rows={repr(self._tile_spec_rows)}, warp_rows={repr(self._warp_rows)}, graphic_layers={repr(self._graphic_layers)}, signs={repr(self._signs)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CitizenOpenClientPacket(Packet):
    """
    Talking to a citizenship NPC
    """
    _byte_size: int = 0
    _npc_index: int

    def __init__(self, *, npc_index: int):
        """
        Create a new instance of CitizenOpenClientPacket.

        Args:
            npc_index (int): (Value range is 0-64008.)
        """
        self._npc_index = npc_index

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def npc_index(self) -> int:
        return self._npc_index

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Citizen

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Open

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        CitizenOpenClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "CitizenOpenClientPacket") -> None:
        """
        Serializes an instance of `CitizenOpenClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CitizenOpenClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._npc_index is None:
                raise SerializationError("npc_index must be provided.")
            writer.add_short(data._npc_index)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CitizenOpenClientPacket":
        """
        Deserializes an instance of `CitizenOpenClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CitizenOpenClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            npc_index = reader.get_short()
            result = CitizenOpenClientPacket(npc_index=npc_index)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CitizenOpenClientPacket(byte_size={repr(self._byte_size)}, npc_index={repr(self._npc_index)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from typing import cast
from typing import Optional
from collections.abc import Iterable
from .npc_update_position import NpcUpdatePosition
from .npc_update_chat import NpcUpdateChat
from .npc_update_attack import NpcUpdateAttack
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class NpcPlayerServerPacket(Packet):
    """
    Main NPC update message
    """
    _byte_size: int = 0
    _positions: tuple[NpcUpdatePosition, ...]
    _attacks: tuple[NpcUpdateAttack, ...]
    _chats: tuple[NpcUpdateChat, ...]
    _hp: Optional[int]
    _tp: Optional[int]

    def __init__(self, *, positions: Iterable[NpcUpdatePosition], attacks: Iterable[NpcUpdateAttack], chats: Iterable[NpcUpdateChat], hp: Optional[int] = None, tp: Optional[int] = None):
        """
        Create a new instance of NpcPlayerServerPacket.

        Args:
            positions (Iterable[NpcUpdatePosition]): 
            attacks (Iterable[NpcUpdateAttack]): 
            chats (Iterable[NpcUpdateChat]): 
            hp (Optional[int]): (Value range is 0-64008.)
            tp (Optional[int]): (Value range is 0-64008.)
        """
        self._positions = tuple(positions)
        self._attacks = tuple(attacks)
        self._chats = tuple(chats)
        self._hp = hp
        self._tp = tp

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def positions(self) -> tuple[NpcUpdatePosition, ...]:
        return self._positions

    @property
    def attacks(self) -> tuple[NpcUpdateAttack, ...]:
        return self._attacks

    @property
    def chats(self) -> tuple[NpcUpdateChat, ...]:
        return self._chats

    @property
    def hp(self) -> Optional[int]:
        return self._hp

    @property
    def tp(self) -> Optional[int]:
        return self._tp

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.Npc

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Player

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        NpcPlayerServerPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "NpcPlayerServerPacket") -> None:
        """
        Serializes an instance of `NpcPlayerServerPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (NpcPlayerServerPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            writer.string_sanitization_mode = True
            if data._positions is None:
                raise SerializationError("positions must be provided.")
            for i in range(len(data._positions)):
                NpcUpdatePosition.serialize(writer, data._positions[i])
            writer.add_byte(0xFF)
            if data._attacks is None:
                raise SerializationError("attacks must be provided.")
            for i in range(len(data._attacks)):
                NpcUpdateAttack.serialize(writer, data._attacks[i])
            writer.add_byte(0xFF)
            if data._chats is None:
                raise SerializationError("chats must be provided.")
            for i in range(len(data._chats)):
                NpcUpdateChat.serialize(writer, data._chats[i])
            writer.add_byte(0xFF)
            reached_missing_optional = data._hp is None
            if not reached_missing_optional:
                writer.add_short(cast(int, data._hp))
            reached_missing_optional = reached_missing_optional or data._tp is None
            if not reached_missing_optional:
                writer.add_short(cast(int, data._tp))
            writer.string_sanitization_mode = False
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "NpcPlayerServerPacket":
        """
        Deserializes an instance of `NpcPlayerServerPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            NpcPlayerServerPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            reader.chunked_reading_mode = True
            positions_length = int(reader.remaining / 4)
            positions = []
            for i in range(positions_length):
                positions.append(NpcUpdatePosition.deserialize(reader))
            reader.next_chunk()
            attacks_length = int(reader.remaining / 9)
            attacks = []
            for i in range(attacks_length):
                attacks.append(NpcUpdateAttack.deserialize(reader))
            reader.next_chunk()
            chats = []
            while reader.remaining > 0:
                chats.append(NpcUpdateChat.deserialize(reader))
            reader.next_chunk()
            hp: Optional[int] = None
            if reader.remaining > 0:
                hp = reader.get_short()
            tp: Optional[int] = None
            if reader.remaining > 0:
                tp = reader.get_short()
            reader.chunked_reading_mode = False
            result = NpcPlayerServerPacket(positions=positions, attacks=attacks, chats=chats, hp=hp, tp=tp)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"NpcPlayerServerPacket(byte_size={repr(self._byte_size)}, positions={repr(self._positions)}, attacks={repr(self._attacks)}, chats={repr(self._chats)}, hp={repr(self._hp)}, tp={repr(self._tp)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from __future__ import annotations
from collections.abc import Iterable
from ..packet_family import PacketFamily
from ..packet_action import PacketAction
from ....serialization_error import SerializationError
from ....net.packet import Packet
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class NpcRangeRequestClientPacket(Packet):
    """
    Requesting info about nearby NPCs
    """
    _byte_size: int = 0
    _npc_indexes_length: int
    _npc_indexes: tuple[int, ...]

    def __init__(self, *, npc_indexes: Iterable[int]):
        """
        Create a new instance of NpcRangeRequestClientPacket.

        Args:
            npc_indexes (Iterable[int]): (Length must be 252 or less.) (Element value range is 0-252.)
        """
        self._npc_indexes = tuple(npc_indexes)
        self._npc_indexes_length = len(self._npc_indexes)

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def npc_indexes(self) -> tuple[int, ...]:
        return self._npc_indexes

    @staticmethod
    def family() -> PacketFamily:
        """
        Returns the packet family associated with this packet.

        Returns:
            PacketFamily: The packet family associated with this packet.
        """
        return PacketFamily.NpcRange

    @staticmethod
    def action() -> PacketAction:
        """
        Returns the packet action associated with this packet.

        Returns:
            PacketAction: The packet action associated with this packet.
        """
        return PacketAction.Request

    def write(self, writer):
        """
        Serializes and writes this packet to the provided EoWriter.

        Args:
            writer (EoWriter): the writer that this packet will be written to.
        """
        NpcRangeRequestClientPacket.serialize(writer, self)

    @staticmethod
    def serialize(writer: EoWriter, data: "NpcRangeRequestClientPacket") -> None:
        """
        Serializes an instance of `NpcRangeRequestClientPacket` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (NpcRangeRequestClientPacket): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._npc_indexes_length is None:
                raise SerializationError("npc_indexes_length must be provided.")
            writer.add_char(data._npc_indexes_length)
            writer.add_byte(255)
            if data._npc_indexes is None:
                raise SerializationError("npc_indexes must be provided.")
            if len(data._npc_indexes) > 252:
                raise SerializationError(f"Expected length of npc_indexes to be 252 or less, got {len(data._npc_indexes)}.")
            for i in range(data._npc_indexes_length):
                writer.add_char(data._npc_indexes[i])
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "NpcRangeRequestClientPacket":
        """
        Deserializes an instance of `NpcRangeRequestClientPacket` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            NpcRangeRequestClientPacket: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            npc_indexes_length = reader.get_char()
            reader.get_byte()
            npc_indexes = []
            for i in range(npc_indexes_length):
                npc_indexes.append(reader.get_char())
            result = NpcRangeRequestClientPacket(npc_indexes=npc_indexes)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"NpcRangeRequestClientPacket(byte_size={repr(self._byte_size)}, npc_indexes={repr(self._npc_indexes)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class MarriageReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with MARRIAGE_REPLY packet
    """
    AlreadyMarried = 1
    NotMarried = 2
    Success = 3
    NotEnoughGold = 4
    WrongName = 5
    ServiceBusy = 6
    DivorceNotification = 7



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from .character_secondary_stats_info_lookup import CharacterSecondaryStatsInfoLookup
from .character_elemental_stats import CharacterElementalStats
from .character_base_stats import CharacterBaseStats
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class CharacterStatsInfoLookup:
    """
    Character stats data.
    Sent with character info lookups.
    """
    _byte_size: int = 0
    _hp: int
    _max_hp: int
    _tp: int
    _max_tp: int
    _base_stats: CharacterBaseStats
    _secondary_stats: CharacterSecondaryStatsInfoLookup
    _elemental_stats: CharacterElementalStats

    def __init__(self, *, hp: int, max_hp: int, tp: int, max_tp: int, base_stats: CharacterBaseStats, secondary_stats: CharacterSecondaryStatsInfoLookup, elemental_stats: CharacterElementalStats):
        """
        Create a new instance of CharacterStatsInfoLookup.

        Args:
            hp (int): (Value range is 0-64008.)
            max_hp (int): (Value range is 0-64008.)
            tp (int): (Value range is 0-64008.)
            max_tp (int): (Value range is 0-64008.)
            base_stats (CharacterBaseStats): 
            secondary_stats (CharacterSecondaryStatsInfoLookup): 
            elemental_stats (CharacterElementalStats): 
        """
        self._hp = hp
        self._max_hp = max_hp
        self._tp = tp
        self._max_tp = max_tp
        self._base_stats = base_stats
        self._secondary_stats = secondary_stats
        self._elemental_stats = elemental_stats

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def hp(self) -> int:
        return self._hp

    @property
    def max_hp(self) -> int:
        return self._max_hp

    @property
    def tp(self) -> int:
        return self._tp

    @property
    def max_tp(self) -> int:
        return self._max_tp

    @property
    def base_stats(self) -> CharacterBaseStats:
        return self._base_stats

    @property
    def secondary_stats(self) -> CharacterSecondaryStatsInfoLookup:
        return self._secondary_stats

    @property
    def elemental_stats(self) -> CharacterElementalStats:
        return self._elemental_stats

    @staticmethod
    def serialize(writer: EoWriter, data: "CharacterStatsInfoLookup") -> None:
        """
        Serializes an instance of `CharacterStatsInfoLookup` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (CharacterStatsInfoLookup): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._hp is None:
                raise SerializationError("hp must be provided.")
            writer.add_short(data._hp)
            if data._max_hp is None:
                raise SerializationError("max_hp must be provided.")
            writer.add_short(data._max_hp)
            if data._tp is None:
                raise SerializationError("tp must be provided.")
            writer.add_short(data._tp)
            if data._max_tp is None:
                raise SerializationError("max_tp must be provided.")
            writer.add_short(data._max_tp)
            if data._base_stats is None:
                raise SerializationError("base_stats must be provided.")
            CharacterBaseStats.serialize(writer, data._base_stats)
            if data._secondary_stats is None:
                raise SerializationError("secondary_stats must be provided.")
            CharacterSecondaryStatsInfoLookup.serialize(writer, data._secondary_stats)
            if data._elemental_stats is None:
                raise SerializationError("elemental_stats must be provided.")
            CharacterElementalStats.serialize(writer, data._elemental_stats)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "CharacterStatsInfoLookup":
        """
        Deserializes an instance of `CharacterStatsInfoLookup` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            CharacterStatsInfoLookup: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            hp = reader.get_short()
            max_hp = reader.get_short()
            tp = reader.get_short()
            max_tp = reader.get_short()
            base_stats = CharacterBaseStats.deserialize(reader)
            secondary_stats = CharacterSecondaryStatsInfoLookup.deserialize(reader)
            elemental_stats = CharacterElementalStats.deserialize(reader)
            result = CharacterStatsInfoLookup(hp=hp, max_hp=max_hp, tp=tp, max_tp=max_tp, base_stats=base_stats, secondary_stats=secondary_stats, elemental_stats=elemental_stats)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"CharacterStatsInfoLookup(byte_size={repr(self._byte_size)}, hp={repr(self._hp)}, max_hp={repr(self._max_hp)}, tp={repr(self._tp)}, max_tp={repr(self._max_tp)}, base_stats={repr(self._base_stats)}, secondary_stats={repr(self._secondary_stats)}, elemental_stats={repr(self._elemental_stats)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class LoginReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with LOGIN_REPLY packet.
    Indicates the result of a login attempt.
    """
    WrongUser = 1
    WrongUserPassword = 2
    Ok = 3
    Banned = 4
    """
    The official client won't display a message until the connection from the server is closed
    """
    LoggedIn = 5
    Busy = 6
    """
    The official client won't display a message until the connection from the server is closed
    """



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ...direction import Direction
from ...coords import Coords
from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class NpcKilledData:
    """
    Information about an NPC that has been killed
    """
    _byte_size: int = 0
    _killer_id: int
    _killer_direction: Direction
    _npc_index: int
    _drop_index: int
    _drop_id: int
    _drop_coords: Coords
    _drop_amount: int
    _damage: int

    def __init__(self, *, killer_id: int, killer_direction: Direction, npc_index: int, drop_index: int, drop_id: int, drop_coords: Coords, drop_amount: int, damage: int):
        """
        Create a new instance of NpcKilledData.

        Args:
            killer_id (int): (Value range is 0-64008.)
            killer_direction (Direction): 
            npc_index (int): (Value range is 0-64008.)
            drop_index (int): (Value range is 0-64008.)
            drop_id (int): (Value range is 0-64008.)
            drop_coords (Coords): 
            drop_amount (int): (Value range is 0-4097152080.)
            damage (int): (Value range is 0-16194276.)
        """
        self._killer_id = killer_id
        self._killer_direction = killer_direction
        self._npc_index = npc_index
        self._drop_index = drop_index
        self._drop_id = drop_id
        self._drop_coords = drop_coords
        self._drop_amount = drop_amount
        self._damage = damage

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def killer_id(self) -> int:
        return self._killer_id

    @property
    def killer_direction(self) -> Direction:
        return self._killer_direction

    @property
    def npc_index(self) -> int:
        return self._npc_index

    @property
    def drop_index(self) -> int:
        return self._drop_index

    @property
    def drop_id(self) -> int:
        return self._drop_id

    @property
    def drop_coords(self) -> Coords:
        return self._drop_coords

    @property
    def drop_amount(self) -> int:
        return self._drop_amount

    @property
    def damage(self) -> int:
        return self._damage

    @staticmethod
    def serialize(writer: EoWriter, data: "NpcKilledData") -> None:
        """
        Serializes an instance of `NpcKilledData` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (NpcKilledData): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._killer_id is None:
                raise SerializationError("killer_id must be provided.")
            writer.add_short(data._killer_id)
            if data._killer_direction is None:
                raise SerializationError("killer_direction must be provided.")
            writer.add_char(int(data._killer_direction))
            if data._npc_index is None:
                raise SerializationError("npc_index must be provided.")
            writer.add_short(data._npc_index)
            if data._drop_index is None:
                raise SerializationError("drop_index must be provided.")
            writer.add_short(data._drop_index)
            if data._drop_id is None:
                raise SerializationError("drop_id must be provided.")
            writer.add_short(data._drop_id)
            if data._drop_coords is None:
                raise SerializationError("drop_coords must be provided.")
            Coords.serialize(writer, data._drop_coords)
            if data._drop_amount is None:
                raise SerializationError("drop_amount must be provided.")
            writer.add_int(data._drop_amount)
            if data._damage is None:
                raise SerializationError("damage must be provided.")
            writer.add_three(data._damage)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "NpcKilledData":
        """
        Deserializes an instance of `NpcKilledData` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            NpcKilledData: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            killer_id = reader.get_short()
            killer_direction = Direction(reader.get_char())
            npc_index = reader.get_short()
            drop_index = reader.get_short()
            drop_id = reader.get_short()
            drop_coords = Coords.deserialize(reader)
            drop_amount = reader.get_int()
            damage = reader.get_three()
            result = NpcKilledData(killer_id=killer_id, killer_direction=killer_direction, npc_index=npc_index, drop_index=drop_index, drop_id=drop_id, drop_coords=drop_coords, drop_amount=drop_amount, damage=damage)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"NpcKilledData(byte_size={repr(self._byte_size)}, killer_id={repr(self._killer_id)}, killer_direction={repr(self._killer_direction)}, npc_index={repr(self._npc_index)}, drop_index={repr(self._drop_index)}, drop_id={repr(self._drop_id)}, drop_coords={repr(self._drop_coords)}, drop_amount={repr(self._drop_amount)}, damage={repr(self._damage)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from ....serialization_error import SerializationError
from .....data.eo_writer import EoWriter
from .....data.eo_reader import EoReader

class ShopCraftIngredientRecord:
    """
    Record of an ingredient for crafting an item in a shop
    """
    _byte_size: int = 0
    _item_id: int
    _amount: int

    def __init__(self, *, item_id: int, amount: int):
        """
        Create a new instance of ShopCraftIngredientRecord.

        Args:
            item_id (int): Item ID of the craft ingredient, or 0 if the ingredient is not present (Value range is 0-64008.)
            amount (int): (Value range is 0-252.)
        """
        self._item_id = item_id
        self._amount = amount

    @property
    def byte_size(self) -> int:
        """
        Returns the size of the data that this was deserialized from.

        Returns:
            int: The size of the data that this was deserialized from.
        """
        return self._byte_size

    @property
    def item_id(self) -> int:
        """
        Item ID of the craft ingredient, or 0 if the ingredient is not present
        """
        return self._item_id

    @property
    def amount(self) -> int:
        return self._amount

    @staticmethod
    def serialize(writer: EoWriter, data: "ShopCraftIngredientRecord") -> None:
        """
        Serializes an instance of `ShopCraftIngredientRecord` to the provided `EoWriter`.

        Args:
            writer (EoWriter): The writer that the data will be serialized to.
            data (ShopCraftIngredientRecord): The data to serialize.
        """
        old_string_sanitization_mode: bool = writer.string_sanitization_mode
        try:
            if data._item_id is None:
                raise SerializationError("item_id must be provided.")
            writer.add_short(data._item_id)
            if data._amount is None:
                raise SerializationError("amount must be provided.")
            writer.add_char(data._amount)
        finally:
            writer.string_sanitization_mode = old_string_sanitization_mode

    @staticmethod
    def deserialize(reader: EoReader) -> "ShopCraftIngredientRecord":
        """
        Deserializes an instance of `ShopCraftIngredientRecord` from the provided `EoReader`.

        Args:
            reader (EoReader): The writer that the data will be serialized to.

        Returns:
            ShopCraftIngredientRecord: The data to serialize.
        """
        old_chunked_reading_mode: bool = reader.chunked_reading_mode
        try:
            reader_start_position: int = reader.position
            item_id = reader.get_short()
            amount = reader.get_char()
            result = ShopCraftIngredientRecord(item_id=item_id, amount=amount)
            result._byte_size = reader.position - reader_start_position
            return result
        finally:
            reader.chunked_reading_mode = old_chunked_reading_mode

    def __repr__(self):
        return f"ShopCraftIngredientRecord(byte_size={repr(self._byte_size)}, item_id={repr(self._item_id)}, amount={repr(self._amount)})"



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ...protocol_enum_meta import ProtocolEnumMeta

class ItemSize(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Size of an item in the inventory
    """
    Size1x1 = 0
    Size1x2 = 1
    Size1x3 = 2
    Size1x4 = 3
    Size2x1 = 4
    Size2x2 = 5
    Size2x3 = 6
    Size2x4 = 7



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class GuildInfoType(IntEnum, metaclass=ProtocolEnumMeta):
    """
    The type of guild info being interacted with
    """
    Description = 1
    Ranks = 2
    Bank = 3



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

from enum import IntEnum
from ....protocol_enum_meta import ProtocolEnumMeta

class CharacterReply(IntEnum, metaclass=ProtocolEnumMeta):
    """
    Reply code sent with CHARACTER_REPLY packet
    """
    Exists = 1
    Full = 2
    """
    Only sent in reply to Character_Create packets.
    Displays the same message as CharacterReply.Full3 in the official client.
    """
    Full3 = 3
    """
    Only sent in reply to Character_Request packets.
    Displays the same message as CharacterReply.Full in the official client.
    """
    NotApproved = 4
    Ok = 5
    Deleted = 6



# Generated from the eo-protocol XML specification.
#
# This file should not be modified.
# Changes will be lost when code is regenerated.

"""
Data structures generated from the [eo-protocol](https://github.com/cirras/eo-protocol/tree/master/xml/pub/protocol.xml){target="_blank"} XML specification.

Warning:
  - This subpackage should not be directly imported. 
  - Instead, import [eolib.protocol.pub][] (or the top-level `eolib`).
"""

from .skill_type import *
from .skill_target_type import *
from .skill_target_restrict import *
from .skill_nature import *
from .npc_type import *
from .item_type import *
from .item_subtype import *
from .item_special import *
from .item_size import *
from .esf_record import *
from .esf import *
from .enf_record import *
from .enf import *
from .element import *
from .eif_record import *
from .eif import *
from .ecf_record import *
from .ecf import *
